#!/usr/bin/env python
# crate_anon/tools/postcodes.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Code-Point Open, CSV, GB
- https://www.ordnancesurvey.co.uk/business-and-government/products/opendata-products.html
- https://www.ordnancesurvey.co.uk/business-and-government/products/code-point-open.html
- https://www.ordnancesurvey.co.uk/opendatadownload/products.html
- http://www.nationalarchives.gov.uk/doc/open-government-licence/version/3/

Office for National Statistics Postcode Database (ONSPD):
- https://geoportal.statistics.gov.uk/geoportal/catalog/content/filelist.page
- e.g. ONSPD_MAY_2016_csv.zip
- http://www.ons.gov.uk/methodology/geography/licences

Background:
- OA = Output Area
    ... smallest: >=40 households, >=100 people
    ... 181,408 OAs in England & Wales
- LSOA = Lower Layer Super Output Area
    ... 34,753 LSOAs in England & Wales
- MSOA = Middle Layer Super Output Area
    ... 7,201 MSOAs in England & Wales
- WZ = Workplace Zone
    - https://www.ons.gov.uk/methodology/geography/ukgeographies/censusgeography#workplace-zone-wz
- https://www.ons.gov.uk/methodology/geography/ukgeographies/censusgeography#output-area-oa
"""  # noqa

import argparse
import csv
import datetime
import logging
import os
import sys
# import textwrap
from typing import Any, Dict, Generator, Iterable, List, TextIO

import openpyxl
from openpyxl.cell.cell import Cell
import prettytable
from sqlalchemy import (
    Column,
    create_engine,
    Date,
    Integer,
    MetaData,
    Numeric,
    String,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session
import xlrd

from cardinal_pythonlib.rnc_extract_text import docx_process_simple_text
from crate_anon.anonymise.constants import CHARSET, TABLE_KWARGS
from crate_anon.common.fileops import find_first
from crate_anon.common.lang import rename_kwarg
from crate_anon.common.logsupport import configure_logger_for_colour

log = logging.getLogger(__name__)
metadata = MetaData()

DEFAULT_ONSPD_DIR = os.path.join(os.path.expanduser("~"), "dev", "onspd")
DEFAULT_REPORT_EVERY = 1000
DEFAULT_COMMIT_EVERY = 10000
YEAR_MONTH_FMT = "%Y%m"

CODE_LEN = 9  # many ONSPD codes have this length
NAME_LEN = 80  # seems about right; a bit more than the length of many


# =============================================================================
# Ancillary functions
# =============================================================================

def convert_date(kwargs: Dict[str, Any], field: str) -> None:
    if field not in kwargs:
        return
    value = kwargs[field]
    if value:
        kwargs[field] = datetime.datetime.strptime(value,
                                                   YEAR_MONTH_FMT)
    else:
        kwargs[field] = None


def convert_int(kwargs: Dict[str, Any], field: str) -> None:
    if field not in kwargs:
        return
    value = kwargs[field]
    if value is None or (isinstance(value, str) and not value.strip()):
        kwargs[field] = None
    else:
        kwargs[field] = int(value)


def convert_float(kwargs: Dict[str, Any], field: str) -> None:
    if field not in kwargs:
        return
    value = kwargs[field]
    if value is None or (isinstance(value, str) and not value.strip()):
        kwargs[field] = None
    else:
        kwargs[field] = float(value)


def values_from_row(row: Iterable[Cell]) -> List[Any]:
    """For openpyxl interface to XLSX files."""
    values = []
    for cell in row:
        values.append(cell.value)
    return values


def commit_and_announce(session: Session) -> None:
    log.info("COMMIT")
    session.commit()


# =============================================================================
# Extend SQLAlchemy Base class
# http://docs.sqlalchemy.org/en/latest/orm/extensions/declarative/mixins.html
# =============================================================================

class ExtendedBase(object):
    __table_args__ = TABLE_KWARGS


Base = declarative_base(metadata=metadata, cls=ExtendedBase)


# =============================================================================
# Models: all postcodes
# =============================================================================

class Postcode(Base):
    """Maps individual postcodes to... lots of things. Large table."""
    __tablename__ = 'postcode'

    pcd_nospace = Column(String(8), primary_key=True,
                         doc="Postcode (no spaces)")
    # ... not in original, but simplifies indexing
    pcd = Column(String(7), index=True, unique=True,
                 doc="Unit postcode (7 characters): 2-4 char outward code, "
                     "left-aligned; 3-char inward code, right-aligned")
    pcd2 = Column(String(8), index=True, unique=True,
                  doc="Unit postcode (8 characters): 2-4 char outward code, "
                      "left-aligned; space; 3-char inward code, right-aligned")
    pcds = Column(String(8), index=True, unique=True,
                  doc="Unit postcode (variable length): 2-4 char outward "
                      "code; space; 3-char inward code")
    dointr = Column(Date, doc="Date of introduction (original format YYYYMM)")
    doterm = Column(Date,
                    doc="Date of termination (original format YYYYMM) or NULL")
    oscty = Column(String(CODE_LEN),
                   doc="County code [FK to county_england_2010.county_code]")
    oslaua = Column(String(CODE_LEN),
                    doc="Local authority district (LUA), unitary  authority "
                        "(UA), metropolitan district (MD), London borough (LB),"
                        " council area (CA), or district council area (DCA) "
                        "[FK to lad_local_authority_district_2016.lad_code]")
    osward = Column(String(CODE_LEN),
                    doc="Electoral ward/division "
                        "[FK e.g. to electoral_ward_2016.ward_code]")
    usertype = Column(Integer, doc="Small (0) or large (1) postcode user")
    oseast1m = Column(Integer,
                      doc="National grid reference Easting, 1m resolution")
    osnrth1m = Column(Integer,
                      doc="National grid reference Northing, 1m resolution")
    osgrdind = Column(Integer,
                      doc="Grid reference positional quality indicator")
    oshlthau = Column(
        String(CODE_LEN),
        doc="Former (up to 2013) Strategic Health Authority (SHA), Local "
            "Health Board (LHB), Health Board (HB), Health Authority (HA), or "
            "Health & Social Care Board (HSCB) [FK to one of: "
            "sha_strategic_health_authority_england_2010.sha_code or "
            "sha_strategic_health_authority_england_2004.sha_code; "
            "hb_health_board_n_ireland_2003.hb_code; "
            "hb_health_board_scotland_2014.hb_code; "
            "hscb_health_social_care_board_n_ireland_2010.hscb_code; "
            "lhb_local_health_board_wales_2014.lhb_code or "
            "lhb_local_health_board_wales_2006.lhb_code]")
    hro = Column(String(CODE_LEN),
                 doc="Pan SHA [FK to "
                     "psha_pan_strategic_health_authority_aka_hro_england_2010"
                     ".psha_code]")
    ctry = Column(String(CODE_LEN),
                  doc="Country of the UK [England, Scotland, Wales, "
                      "Northern Ireland] [FK to country_2012.country_code]")
    gor = Column(String(CODE_LEN),
                 doc="Region (former GOR) [GOR = Government Office Regions] "
                     "[FK to gor_govt_office_region_england_2010.gor_code]")
    streg = Column(Integer, doc="Standard (Statistical) Region (SSR) [FK to "
                                "ssr_standard_statistical_region_england_1995."
                                "ssr_code]")
    pcon = Column(String(CODE_LEN),
                  doc="Westminster parliamentary constituency [FK to "
                      "pcon_westminster_parliamentary_constituency_2014."
                      "pcon_code]")
    eer = Column(String(CODE_LEN),
                 doc="European Electoral Region (EER) [FK to "
                     "eer_european_electoral_region_2010.eer_code]")
    teclec = Column(
        String(CODE_LEN),
        doc="Local Learning and Skills Council (LLSC) / Dept. of Children, "
            "Education, Lifelong Learning and Skills (DCELLS) / Enterprise "
            "Region (ER) [PROBABLY FK to one of: "
            "dcells_dept_children_wales_2010.dcells_code; "
            "er_enterprise_region_scotland_2010.er_code; "
            "llsc_local_learning_skills_council_england_2010.llsc_code]")
    ttwa = Column(String(CODE_LEN),
                  doc="Travel to Work Area (TTWA) [FK to "
                      "ttwa_travel_to_work_area_2011.ttwa_code]")
    pct = Column(
        String(CODE_LEN),
        doc="Primary Care Trust (PCT) / Care Trust / Care Trust Plus (CT) / "
            "Local Health Board (LHB) / Community Health Partnership (CHP) / "
            "Local Commissioning Group (LCG) / Primary Healthcare Directorate "
            "(PHD) [FK to one of: "
            "pct_primary_care_trust_organization_england_2011.pct_code; "
            "chp_community_health_partnership_scotland_2012.chp_code; "
            "lcg_local_commissioning_group_n_ireland_2010.lcg_code; "
            "lhb_local_health_board_wales_2014.lhb_code]")
    nuts = Column(String(10),
                  doc="LAU2 areas [European Union spatial regions; Local "
                      "Adminstrative Unit, level 2] / Nomenclature of Units "
                      "for Territorial Statistics (NUTS) [FK to "
                      "lau_eu_local_administrative_unit_2015.lau2_code]")
    psed = Column(String(8),
                  doc="1991 Census Enumeration District (ED) (as OGSS code) "
                      "[POSSIBLY FK to district_england_wales_1991."
                      "district_code_ogss]")
    cened = Column(String(8),
                   doc="1991 Census Enumeration District (ED) (as Census code)"
                       " [POSSIBLY FK to district_england_wales_1991."
                       "district_code_census]")
    edind = Column(Integer, doc="ED positional quality indicator (PQI)")
    oshaprev = Column(
        String(3),
        doc="Previous Strategic Health Authority (SHA) / Health Board (HB) / "
            "Health Authority (HA) / Health and Social Services Board (HSSB) "
            "[FK to one of: "
            "sha_strategic_health_authority_england_2010.sha_code or "
            "sha_strategic_health_authority_england_2004.sha_code]")
    lea = Column(String(3),
                 doc="Previous Local Education Authority (LEA) / Education "
                     "and Library Board (ELB) [PROBABLY FK to "
                     "lea_local_education_authority_2009.lea_code]")
    oldha = Column(String(3),
                   doc="Old-style health authority (prior to 2002 "
                       "England, 2003 Wales [FK to "
                       "ha_health_authority_2001.ha_code]")
    wardc91 = Column(String(6),
                     doc="1991 administrative and electoral area (as Census "
                         "code) [FK to electoral_ward_1991.ward_code_census]")
    wardo91 = Column(String(6),
                     doc="1991 administrative and electoral area (as OGSS "
                         "code) [FK to electoral_ward_1991.ward_code_ogss]")
    ward98 = Column(String(6),
                    doc="1998 administrative and electoral area [FK to "
                        "electoral_ward_1998.ward_code]")
    statsward = Column(String(6),
                       doc="2005 'statistical' ward [?FK to "
                           "electoral_ward_2005.ward_code]")
    oa01 = Column(String(10), doc="2001 Census Output Area (OA). (There are "
                                  "about 222,000, so ~300 population?)")
    casward = Column(String(6),
                     doc="Census Area Statistics (CAS) ward [PROBABLY FK to "
                         "cas_ward_2003.cas_ward_code]")
    park = Column(String(CODE_LEN), doc="National park [FK to "
                                        "park_national_park_2010.park_code]")
    lsoa01 = Column(
        String(CODE_LEN),
        doc="2001 Census Lower Layer Super Output Area (LSOA) [England & "
            "Wales, ~1,500 population] / Data Zone (DZ) [Scotland] / Super "
            "Output Area (SOA) [FK to one of: "
            "lsoa_lower_layer_super_output_area_england_wales_2004.lsoa_code; "
            "lsoa_lower_layer_super_output_area_n_ireland_2005.lsoa_code]")
    msoa01 = Column(
        String(CODE_LEN),
        doc="2001 Census Middle Layer Super Output Area (MSOA) [England & "
            "Wales, ~7,200 population] / Intermediate Zone (IZ) [Scotland] "
            "[FK to one of: "
            "msoa_middle_layer_super_output_area_england_wales_2004.msoa_code;"
            " iz_intermediate_zone_scotland_2005.iz_code]")
    ur01ind = Column(String(1),
                     doc="2001 Census urban/rural indicator [numeric in "
                         "England/Wales/Scotland; letters in N. Ireland]")
    oac01 = Column(String(3),
                   doc="2001 Census Output Area classification (OAC)"
                       "[POSSIBLY FK to output_area_classification_2011."
                       "subgroup_code]")
    oldpct = Column(
        String(5),
        doc="Old (pre-Oct 2006) Primary Care Trust (PCT) / Local Health Board "
            "(LHB) / Care Trust (CT) [FK to one of: "
            "pct_primary_care_trust_organization_england_2005.pct_code]")
    oa11 = Column(String(CODE_LEN),
                  doc="2011 Census Output Area (OA) [England, Wales, Scotland;"
                      " ~100-625 population] / Small Area (SA) [N. Ireland]")
    lsoa11 = Column(
        String(CODE_LEN),
        doc="2011 Census Lower Layer Super Output Area (LSOA) [England & "
            "Wales, ~1,500 population] / Data Zone (DZ) [Scotland] / Super "
            "Output Area (SOA) [FK to one of: "
            "lsoa_lower_layer_super_output_area_england_wales_2011.lsoa_code;"
            " dz_datazone_scotland_2011.dz_code]")
    msoa11 = Column(
        String(CODE_LEN),
        doc="2011 Census Middle Layer Super Output Area (MSOA) [England & "
            "Wales, ~7,200 population] / Intermediate Zone (IZ) [Scotland] "
            "[FK to one of: "
            "msoa_middle_layer_super_output_area_england_wales_2011.msoa_code;"
            " iz_intermediate_zone_scotland_2011.iz_code]")
    parish = Column(String(CODE_LEN),
                    doc="Parish/community [FK to "
                        "parish_lad_england_wales_2014.parish_code; "
                        "ncp_non_civil_parish_2014.ncp_code]")
    wz11 = Column(String(CODE_LEN), doc="2011 Census Workplace Zone (WZ)")
    ccg = Column(
        String(CODE_LEN),
        doc="Clinical Commissioning Group (CCG) / Local Health Board (LHB) / "
            "Community Health Partnership (CHP) / Local Commissioning Group "
            "(LCG) / Primary Healthcare Directorate (PHD) [FK to one of: "
            "ccg_clinical_commissioning_group_england_2015."
            "ccg_ons_code, lhb_local_health_board_wales_2014.lhb_code]")
    bua11 = Column(String(CODE_LEN),
                   doc="Built-up Area (BUA) [FK to "
                       "bua_built_up_area_england_wales_2013.bua_code]")
    buasd11 = Column(
        String(CODE_LEN),
        doc="Built-up Area Sub-division (BUASD) [FK to "
            "buasd_built_up_area_subdivision_england_wales_2013.buas_code]")
    ru11ind = Column(String(2), doc="2011 Census rural-urban classification")
    oac11 = Column(String(3),
                   doc="2011 Census Output Area classification (OAC) [FK to "
                       "output_area_classification_2011.subgroup_code]")
    lat = Column(Numeric(precision=9, scale=6),
                 doc="Latitude (degrees, 6dp)")
    long = Column(Numeric(precision=9, scale=6),
                  doc="Longitude (degrees, 6dp)")
    lep1 = Column(
        String(CODE_LEN),
        doc="Local Enterprise Partnership (LEP) - first instance [FK to "
            "lep_local_enterprise_partnership_england_2013.lep1_code]")
    lep2 = Column(
        String(CODE_LEN),
        doc="Local Enterprise Partnership (LEP) - second instance [FK to "
            "lep_local_enterprise_partnership_england_2013.lep1_code]")
    pfa = Column(String(CODE_LEN),
                 doc="Police Force Area (PFA) [FK to "
                     "pfa_police_force_area_2015.pfa_code]")
    imd = Column(
        Integer,
        doc="Index of Multiple Deprivation (IMD) [rank of LSOA/DZ, where 1 is "
            "the most deprived, within each country] [FK to one of: "
            "imd_index_multiple_deprivation_england_2015.imd_rank; "
            "imd_index_multiple_deprivation_n_ireland_2010.imd_rank; "
            "imd_index_multiple_deprivation_scotland_2012.imd_rank; "
            "imd_index_multiple_deprivation_wales_2014.imd_rank]")

    def __init__(self, **kwargs: Any) -> None:
        convert_date(kwargs, 'dointr')
        convert_date(kwargs, 'doterm')
        convert_int(kwargs, 'usertype')
        convert_int(kwargs, 'oseast1m')
        convert_int(kwargs, 'osnrth1m')
        convert_int(kwargs, 'osgrdind')
        convert_int(kwargs, 'streg')
        convert_int(kwargs, 'edind')
        convert_int(kwargs, 'imd')
        kwargs['pcd_nospace'] = kwargs['pcd'].replace(" ", "")
        super().__init__(**kwargs)


# =============================================================================
# Models: core lookup tables
# =============================================================================

class OAClassification(Base):
    __filename__ = "2011 Census Output Area Classification Names and Codes " \
                   "UK.xlsx"
    __tablename__ = "output_area_classification_2011"

    supergroup_code = Column(String(1))
    supergroup_desc = Column(String(35))
    group_desc = Column(String(40))
    group_code = Column(String(2))
    subgroup_desc = Column(String(60))
    subgroup_code = Column(String(3), primary_key=True)

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'Supergroup', 'supergroup_desc')
        rename_kwarg(kwargs, 'Group', 'group_desc')
        rename_kwarg(kwargs, 'Subgroup', 'subgroup_desc')
        kwargs['supergroup_code'] = kwargs['supergroup_desc'][0:1]
        kwargs['group_code'] = kwargs['group_desc'][0:2]
        kwargs['subgroup_code'] = kwargs['subgroup_desc'][0:3]
        super().__init__(**kwargs)


class BUA(Base):
    __filename__ = "BUA_names and codes EW as at 12_13.xlsx"
    __tablename__ = "bua_built_up_area_england_wales_2013"

    bua_code = Column(String(CODE_LEN), primary_key=True)
    bua_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'BUA13CD', 'bua_code')
        rename_kwarg(kwargs, 'BUA13NM', 'bua_name')
        super().__init__(**kwargs)


class BUASD(Base):
    __filename__ = "BUASD_names and codes EW as at 12_13.xlsx"
    __tablename__ = "buasd_built_up_area_subdivision_england_wales_2013"

    buasd_code = Column(String(CODE_LEN), primary_key=True)
    buasd_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'BUASD13CD', 'buasd_code')
        rename_kwarg(kwargs, 'BUASD13NM', 'buasd_name')
        super().__init__(**kwargs)


class CASWard(Base):
    __filename__ = "CAS ward names and codes UK as at 01_03.xlsx"
    __tablename__ = "cas_ward_2003"

    cas_ward_code = Column(String(CODE_LEN), primary_key=True)
    cas_ward_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'WDCAS03CD', 'cas_ward_code')
        rename_kwarg(kwargs, 'WDCAS03NM', 'cas_ward_name')
        super().__init__(**kwargs)


class CCG(Base):
    __filename__ = "CCG names and codes EN as at 07_15.xlsx"
    __tablename__ = "ccg_clinical_commissioning_group_england_2015"

    ccg_ons_code = Column(String(CODE_LEN), primary_key=True)
    ccg_ccg_code = Column(String(3))
    ccg_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'CCG15CD', 'ccg_ons_code')
        rename_kwarg(kwargs, 'CCG15CDH', 'ccg_ccg_code')
        rename_kwarg(kwargs, 'CCG15NM', 'ccg_name')
        super().__init__(**kwargs)


class CHP(Base):
    __filename__ = "CHP names and codes SC as at 04_12.xlsx"
    __tablename__ = "chp_community_health_partnership_scotland_2012"

    chp_code = Column(String(CODE_LEN), primary_key=True)
    chp_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'CHP12CD', 'chp_code')
        rename_kwarg(kwargs, 'CHP12NM', 'chp_name')
        super().__init__(**kwargs)


class Country(Base):
    __filename__ = "Country names and codes UK as at 08_12.xls"
    __tablename__ = "country_2012"

    country_code = Column(String(CODE_LEN), primary_key=True)
    country_name = Column(String(NAME_LEN))
    country_name_welsh = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'CTRY12CD', 'country_code')
        rename_kwarg(kwargs, 'CTRY12NM', 'country_name')
        rename_kwarg(kwargs, 'CTRY12NMW', 'country_name_welsh')
        super().__init__(**kwargs)


class County1991(Base):
    __filename__ = "County names and codes EW as at 21_04_91.xls"
    __tablename__ = "county_england_wales_1991"
    __headings__ = ["county_code_census", "county_code_ogss",
                    "county_code_ons", "county_name"]
    # __debug_content__ = True

    county_code_census = Column(String(2))
    county_code_ogss = Column(String(2))
    county_code_ons = Column(String(2))
    county_name = Column(String(NAME_LEN), primary_key=True)
    # ... no code type is present for all counties

    def __init__(self, **kwargs: Any) -> None:
        # The "00" codes are stored in the spreadsheet as a numeric 0 with
        # leading zero formatting.
        def process_00(field):
            value = kwargs[field]
            if value == '':
                kwargs[field] = None
            else:
                kwargs[field] = "{:0>2}".format(int(value))

        process_00('county_code_census')
        process_00('county_code_ogss')
        process_00('county_code_ons')
        super().__init__(**kwargs)


class County2010(Base):
    __filename__ = "County names and codes EN as at 12_10.xls"
    __tablename__ = "county_england_2010"

    county_code = Column(String(CODE_LEN), primary_key=True)
    county_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'CTY10CD', 'county_code')
        rename_kwarg(kwargs, 'CTY10NM', 'county_name')
        super().__init__(**kwargs)


class Datazone(Base):
    __filename__ = "Datazone (2011) names and codes SC as at 11_14.xlsx"
    __tablename__ = "dz_datazone_scotland_2011"

    dz_code = Column(String(CODE_LEN), primary_key=True)
    dz_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'DZ11CD', 'dz_code')
        rename_kwarg(kwargs, 'DZ11NM', 'dz_name')
        super().__init__(**kwargs)


class DCELLS(Base):
    __filename__ = "DCELLS names and codes WA as at 12_10.xls"
    __tablename__ = "dcells_dept_children_wales_2010"

    dcells_code = Column(String(CODE_LEN), primary_key=True)
    dcells_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'DCELL10CD', 'dcells_code')
        rename_kwarg(kwargs, 'DCELL10NM', 'dcells_name')
        super().__init__(**kwargs)
        

class District(Base):
    __filename__ = "District names and codes EW as at 21_4_91.xls"
    __headings__ = ["district_name", "district_code_census",
                    "district_code_ogss", "district_code_ons"]
    __tablename__ = "district_england_wales_1991"

    district_name = Column(String(NAME_LEN))
    district_code_census = Column(String(4))
    district_code_ogss = Column(String(4))
    district_code_ons = Column(String(4), primary_key=True)

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)


class EER(Base):
    __filename__ = "EER names and codes UK as at 12_10.xls"
    __tablename__ = "eer_european_electoral_region_2010"

    eer_code = Column(String(CODE_LEN), primary_key=True)
    eer_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'EER10CD', 'eer_code')
        rename_kwarg(kwargs, 'EER10NM', 'eer_name')
        super().__init__(**kwargs)


class EnterpriseRegion(Base):
    __filename__ = "Enterprise Region names and codes SC as at 12_10.xls"
    __tablename__ = "er_enterprise_region_scotland_2010"

    er_code = Column(String(CODE_LEN), primary_key=True)
    er_code_old = Column(String(3))
    er_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'ER10CD', 'er_code')
        rename_kwarg(kwargs, 'ER10CDO', 'er_code_old')
        rename_kwarg(kwargs, 'ER10NM', 'er_name')
        super().__init__(**kwargs)


class HealthAuthority(Base):
    __filename__ = "Health Authority & Health Board names and codes GB as " \
                   "at 12_01.xls"
    __tablename__ = "ha_health_authority_2001"

    ha_code = Column(String(3), primary_key=True)
    ha_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'HA01CD', 'ha_code')
        rename_kwarg(kwargs, 'HA01NM', 'ha_name')
        super().__init__(**kwargs)


class HealthBoardNI(Base):
    __filename__ = "Health Board names and codes NI as at 2003.xls"
    __tablename__ = "hb_health_board_n_ireland_2003"

    hb_code = Column(String(CODE_LEN), primary_key=True)
    hb_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'HA03CD', 'hb_code')
        rename_kwarg(kwargs, 'HA03NM', 'hb_name')
        super().__init__(**kwargs)


class HealthBoardSC(Base):
    __filename__ = "Health Board names and codes SC as at 12_14.xlsx"
    __tablename__ = "hb_health_board_scotland_2014"

    hb_code = Column(String(CODE_LEN), primary_key=True)
    hb_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'HB14CD', 'hb_code')
        rename_kwarg(kwargs, 'HB14NM', 'hb_name')
        super().__init__(**kwargs)


class HSCBNI(Base):
    __filename__ = "HSCB name and code NI as at 12_10.xls"
    __tablename__ = "hscb_health_social_care_board_n_ireland_2010"

    hscb_code = Column(String(5), primary_key=True)
    hscb_code_2 = Column(String(5))
    hscb_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'HSCB10CDO', 'hscb_code')
        rename_kwarg(kwargs, 'HSCB10CD', 'hscb_code_2')
        rename_kwarg(kwargs, 'HSCB10NM', 'hscb_name')
        super().__init__(**kwargs)


class IMDLookupEN(Base):
    __filename__ = "IMD lookup EN as at 12_15.xlsx"
    __tablename__ = "imd_index_multiple_deprivation_england_2015"

    lsoa_code = Column(String(CODE_LEN), primary_key=True)
    lsoa_name = Column(String(NAME_LEN))
    imd_rank = Column(Integer)

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LSOA11CD', 'lsoa_code')
        rename_kwarg(kwargs, 'LSOA11NM', 'lsoa_name')
        rename_kwarg(kwargs, 'IMD15', 'imd_rank')
        convert_int(kwargs, 'imd_rank')
        super().__init__(**kwargs)


class IMDLookupNI(Base):
    __filename__ = "IMD lookup NI as at 12_10.xlsx"
    __tablename__ = "imd_index_multiple_deprivation_n_ireland_2010"

    oa_code = Column(String(10), primary_key=True)
    imd_rank = Column(Integer)

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'OA01CD', 'oa_code')
        rename_kwarg(kwargs, 'IMD10', 'imd_rank')
        convert_int(kwargs, 'imd_rank')
        super().__init__(**kwargs)


class IMDLookupSC(Base):
    __filename__ = "IMD lookup SC as at 12_12.xlsx"
    __tablename__ = "imd_index_multiple_deprivation_scotland_2012"

    dz_code = Column(String(CODE_LEN), primary_key=True)
    imd_rank = Column(Integer)

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'DZ01CD', 'dz_code')
        rename_kwarg(kwargs, 'IMD12', 'imd_rank')
        convert_int(kwargs, 'imd_rank')
        super().__init__(**kwargs)


class IMDLookupWA(Base):
    __filename__ = "IMD lookup WA as at 12_14.xlsx"
    __tablename__ = "imd_index_multiple_deprivation_wales_2014"

    lsoa_code = Column(String(CODE_LEN), primary_key=True)
    lsoa_name = Column(String(NAME_LEN))
    imd_rank = Column(Integer)

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LSOA11CD', 'lsoa_code')
        rename_kwarg(kwargs, 'LSOA11NM', 'lsoa_name')
        rename_kwarg(kwargs, 'IMD14', 'imd_rank')
        convert_int(kwargs, 'imd_rank')
        super().__init__(**kwargs)


class IZ2005(Base):
    __filename__ = "IZ (2001) names and codes SC as at 11_11.xls"
    # definitely 2005, from metadata
    __tablename__ = "iz_intermediate_zone_scotland_2005"

    iz_code = Column(String(CODE_LEN), primary_key=True)
    iz_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'IZ05CD', 'iz_code')
        rename_kwarg(kwargs, 'IZ05NM', 'iz_name')
        super().__init__(**kwargs)


class IZ2011(Base):
    __filename__ = "IZ (2011) names and codes SC as at 11_14.xlsx"
    __tablename__ = "iz_intermediate_zone_scotland_2011"

    iz_code = Column(String(CODE_LEN), primary_key=True)
    iz_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'IZ11CD', 'iz_code')
        rename_kwarg(kwargs, 'IZ11NM', 'iz_name')
        super().__init__(**kwargs)


class LAU(Base):
    __filename__ = "LAU215_LAU115_NUTS315_NUTS215_NUTS115_UK_LUv2.xlsx"
    __tablename__ = "lau_eu_local_administrative_unit_2015"

    lau2_code = Column(String(10), primary_key=True)
    lau2_name = Column(String(NAME_LEN))
    lau1_code = Column(String(CODE_LEN))
    lau1_name = Column(String(NAME_LEN))
    nuts3_code = Column(String(5))
    nuts3_name = Column(String(NAME_LEN))
    nuts2_code = Column(String(4))
    nuts2_name = Column(String(NAME_LEN))
    nuts1_code = Column(String(3))
    nuts1_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LAU215CD', 'lau2_code')
        rename_kwarg(kwargs, 'LAU215NM', 'lau2_name')
        rename_kwarg(kwargs, 'LAU115CD', 'lau1_code')
        rename_kwarg(kwargs, 'LAU115NM', 'lau1_name')
        rename_kwarg(kwargs, 'NUTS315CD', 'nuts3_code')
        rename_kwarg(kwargs, 'NUTS315NM', 'nuts3_name')
        rename_kwarg(kwargs, 'NUTS215CD', 'nuts2_code')
        rename_kwarg(kwargs, 'NUTS215NM', 'nuts2_name')
        rename_kwarg(kwargs, 'NUTS115CD', 'nuts1_code')
        rename_kwarg(kwargs, 'NUTS115NM', 'nuts1_name')
        super().__init__(**kwargs)


class LAD(Base):
    __filename__ = "LA_UA names and codes UK as at 02_16.xlsx"
    __tablename__ = "lad_local_authority_district_2016"

    lad_code = Column(String(CODE_LEN), primary_key=True)
    lad_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LAD16CD', 'lad_code')
        rename_kwarg(kwargs, 'LAD16NM', 'lad_name')
        super().__init__(**kwargs)


class LCG(Base):
    __filename__ = "LCG names and codes NI as at 12_10.xls"
    __tablename__ = "lcg_local_commissioning_group_n_ireland_2010"

    lcg_code = Column(String(CODE_LEN))
    lcg_code_old = Column(String(5), primary_key=True)
    lcg_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LCG10CD', 'lcg_code')
        rename_kwarg(kwargs, 'LCG10CDO', 'lcg_code_old')
        rename_kwarg(kwargs, 'LCG10NM', 'lcg_name')
        super().__init__(**kwargs)


class LEA(Base):
    __filename__ = "LEA and ELB names and codes UK as at 04_09.xls"
    __tablename__ = "lea_local_education_authority_2009"
    # __debug_content__ = True

    lea_code = Column(String(3), primary_key=True)
    lea_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'EDUC09CD', 'lea_code')
        rename_kwarg(kwargs, 'EDUC09NM', 'lea_name')
        super().__init__(**kwargs)


class LEP(Base):
    __filename__ = "LEP names and codes EN as at 12_13.xlsx"
    __tablename__ = "lep_local_enterprise_partnership_england_2013"
    # __debug_content__ = True

    lep1_code = Column(String(CODE_LEN), primary_key=True)
    lep1_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LEP13CD1', 'lep1_code')
        rename_kwarg(kwargs, 'LEP13NM1', 'lep1_name')
        super().__init__(**kwargs)


class LHB2014(Base):
    __filename__ = "LHB names and codes WA as at 12_14.xlsx"
    __tablename__ = "lhb_local_health_board_wales_2014"
    # __debug_content__ = True

    lhb_code = Column(String(CODE_LEN), primary_key=True)
    lhb_name = Column(String(NAME_LEN))
    lhb_name_welsh = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LHB14CD', 'lhb_code')
        rename_kwarg(kwargs, 'LHB14NM', 'lhb_name')
        rename_kwarg(kwargs, 'LHB14NMW', 'lhb_name_welsh')
        super().__init__(**kwargs)


class LHB2006(Base):
    __filename__ = "Local Health Boards names and codes WA as at 06_06.xls"
    __tablename__ = "lhb_local_health_board_wales_2006"

    lhb_code = Column(String(3), primary_key=True)
    lhb_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LHB06CD', 'lhb_code')
        rename_kwarg(kwargs, 'LHB06NM', 'lhb_name')
        super().__init__(**kwargs)


class LLSC(Base):
    __filename__ = "LLSC names and codes EN as at 12_10.xls"
    __tablename__ = "llsc_local_learning_skills_council_england_2010"

    llsc_code = Column(String(CODE_LEN), primary_key=True)
    llsc_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LLSC10CD', 'llsc_code')
        rename_kwarg(kwargs, 'LLSC10NM', 'llsc_name')
        super().__init__(**kwargs)


class LSOAEW2004(Base):
    __filename__ = "LSOA (2001) names and codes EW as at 02_04.xls"
    __tablename__ = "lsoa_lower_layer_super_output_area_england_wales_2004"

    lsoa_code = Column(String(CODE_LEN), primary_key=True)
    lsoa_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LSOA04CD', 'lsoa_code')
        rename_kwarg(kwargs, 'LSOA04NM', 'lsoa_name')
        super().__init__(**kwargs)


class LSOANI2005(Base):
    __filename__ = "LSOA (2001) names and codes NI as at 02_05.xls"
    __tablename__ = "lsoa_lower_layer_super_output_area_n_ireland_2005"

    lsoa_code = Column(String(8), primary_key=True)
    lsoa_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LSOAN05CD', 'lsoa_code')
        rename_kwarg(kwargs, 'LSOAN05NM', 'lsoa_name')
        super().__init__(**kwargs)


class LSOAEW2011(Base):
    __filename__ = "LSOA (2011) names and codes EW as at 12_12.xlsx"
    __tablename__ = "lsoa_lower_layer_super_output_area_england_wales_2011"

    lsoa_code = Column(String(CODE_LEN), primary_key=True)
    lsoa_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LSOA11CD', 'lsoa_code')
        rename_kwarg(kwargs, 'LSOA11NM', 'lsoa_name')
        super().__init__(**kwargs)


class MSOAEW2004(Base):
    __filename__ = "MSOA (2001) names and codes EW as at 08_04.xls"
    __tablename__ = "msoa_middle_layer_super_output_area_england_wales_2004"

    msoa_code = Column(String(CODE_LEN), primary_key=True)
    msoa_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'MSOA04CD', 'msoa_code')
        rename_kwarg(kwargs, 'MSOA04NM', 'msoa_name')
        super().__init__(**kwargs)


class MSOAEW2011(Base):
    __filename__ = "MSOA (2011) names and codes EW as at 12_12.xlsx"
    __tablename__ = "msoa_middle_layer_super_output_area_england_wales_2011"

    msoa_code = Column(String(CODE_LEN), primary_key=True)
    msoa_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'MSOA11CD', 'msoa_code')
        rename_kwarg(kwargs, 'MSOA11NM', 'msoa_name')
        super().__init__(**kwargs)


class NationalPark(Base):
    __filename__ = "National Park names and codes GB as at 10_10.xls"
    __tablename__ = "park_national_park_2010"

    park_code = Column(String(CODE_LEN), primary_key=True)
    park_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'NPARK10CD', 'park_code')
        rename_kwarg(kwargs, 'NPARK10NM', 'park_name')
        super().__init__(**kwargs)


class PanSHA(Base):
    __filename__ = "Pan SHA names and codes EN as at 12_10.xls"
    __tablename__ = "psha_pan_strategic_health_authority_aka_hro_england_2010"

    psha_code = Column(String(CODE_LEN), primary_key=True)
    psha_code_old = Column(String(3))
    psha_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'PSHA10CD', 'psha_code')
        rename_kwarg(kwargs, 'PSHA10CDO', 'psha_code_old')
        rename_kwarg(kwargs, 'PSHA10NM', 'psha_name')
        super().__init__(**kwargs)


class Parish(Base):
    __filename__ = "Parish LAD names and codes EW as at 12_14.xlsx"
    __tablename__ = "parish_lad_england_wales_2014"

    parish_code = Column(String(CODE_LEN), primary_key=True)
    parish_name = Column(String(NAME_LEN))
    district_code = Column(String(CODE_LEN))
    district_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'PAR14CD', 'parish_code')
        rename_kwarg(kwargs, 'PAR14NM', 'parish_name')
        rename_kwarg(kwargs, 'LAD14CD', 'district_code')
        rename_kwarg(kwargs, 'LAD14NM', 'district_name')
        super().__init__(**kwargs)


class PCT2011(Base):
    __filename__ = "PCO names and codes EN as at 04_11.xls"
    __tablename__ = "pct_primary_care_trust_organization_england_2011"

    pct_code = Column(String(CODE_LEN), primary_key=True)
    pct_code_old = Column(String(3))
    pct_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'PCO11CD', 'pct_code')
        rename_kwarg(kwargs, 'PCO11CDO', 'pct_code_old')
        rename_kwarg(kwargs, 'PCO11NM', 'pct_name')
        super().__init__(**kwargs)


class PCT2005(Base):
    __filename__ = "PCO names and codes EN as at 10_05.xls"
    __tablename__ = "pct_primary_care_trust_organization_england_2005"

    pct_code = Column(String(3), primary_key=True)
    pct_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'PCO05CD', 'pct_code')
        rename_kwarg(kwargs, 'PCO05NM', 'pct_name')
        super().__init__(**kwargs)


class PFA(Base):
    __filename__ = "PFA names and codes GB as at 12_15.xlsx"
    __tablename__ = "pfa_police_force_area_2015"

    pfa_code = Column(String(CODE_LEN), primary_key=True)
    pfa_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'PFA15CD', 'pfa_code')
        rename_kwarg(kwargs, 'PFA15NM', 'pfa_name')
        super().__init__(**kwargs)


class GOR(Base):
    __filename__ = "Region (GOR) names and codes EN as at 12_10.xls"
    __tablename__ = "gor_govt_office_region_england_2010"

    gor_code = Column(String(CODE_LEN), primary_key=True)
    gor_name = Column(String(NAME_LEN))
    gor_name_welsh = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'GOR10CD', 'gor_code')
        rename_kwarg(kwargs, 'GOR10NM', 'gor_name')
        rename_kwarg(kwargs, 'GOR10NMW', 'gor_name')
        super().__init__(**kwargs)


class SHA2004(Base):
    __filename__ = "SHA names and codes EN as at 09_02_04.xls"
    __tablename__ = "sha_strategic_health_authority_england_2004"

    sha_code = Column(String(3), primary_key=True)
    sha_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'SHA04CD', 'sha_code')
        rename_kwarg(kwargs, 'SHA04NM', 'sha_name')
        super().__init__(**kwargs)


class SHA2010(Base):
    __filename__ = "SHA names and codes EN as at 12_10.xls"
    __tablename__ = "sha_strategic_health_authority_england_2010"

    sha_code = Column(String(CODE_LEN), primary_key=True)
    sha_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'SHA10CD', 'sha_code')
        rename_kwarg(kwargs, 'SHA10NM', 'sha_name')
        super().__init__(**kwargs)


class SSR(Base):
    __filename__ = "SSR names and codes EN as at 12_05.xls"
    __tablename__ = "ssr_standard_statistical_region_england_1995"

    ssr_code = Column(Integer, primary_key=True)
    ssr_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'SSR95CD', 'ssr_code')
        rename_kwarg(kwargs, 'SSR95NM', 'ssr_name')
        convert_int(kwargs, 'ssr_code')
        super().__init__(**kwargs)


class Ward1991(Base):
    __filename__ = "Ward names and codes UK as at 21_04_91.xls"
    __tablename__ = "electoral_ward_1991"

    ward_code_ons = Column(String(6), primary_key=True)
    ward_code_census = Column(String(6))
    ward_code_ogss = Column(String(6))
    ward_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'WD91CD', 'ward_code_ons')
        rename_kwarg(kwargs, 'WD91CDC', 'ward_code_census')
        rename_kwarg(kwargs, 'WD91CDO', 'ward_code_ogss')
        rename_kwarg(kwargs, 'WD91NM', 'ward_name')
        super().__init__(**kwargs)


class Ward1998(Base):
    __filename__ = "Ward names and codes UK as at 12_98.xls"
    __tablename__ = "electoral_ward_1998"

    ward_code = Column(String(6), primary_key=True)
    ward_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'WD98CD', 'ward_code')
        rename_kwarg(kwargs, 'WD98NM', 'ward_name')
        super().__init__(**kwargs)


class Ward2005(Base):
    __filename__ = "Statistical ward names and codes UK as at 2005.xls"
    __tablename__ = "electoral_ward_2005"

    ward_code = Column(String(6), primary_key=True)
    ward_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'WDSTL05CD', 'ward_code')
        rename_kwarg(kwargs, 'WDSTL05NM', 'ward_name')
        super().__init__(**kwargs)


class Ward2016(Base):
    __filename__ = "Ward names and codes UK as at 05_16.xlsx"
    __tablename__ = "electoral_ward_2016"

    ward_code = Column(String(CODE_LEN), primary_key=True)
    ward_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'WD16CD', 'ward_code')
        rename_kwarg(kwargs, 'WD16NM', 'ward_name')
        super().__init__(**kwargs)


class TTWA(Base):
    __filename__ = "TTWA names and codes UK as at 12_11 v5.xlsx"
    __tablename__ = "ttwa_travel_to_work_area_2011"

    ttwa_code = Column(String(CODE_LEN), primary_key=True)
    ttwa_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'TTWA11CD', 'ttwa_code')
        rename_kwarg(kwargs, 'TTWA11NM', 'ttwa_name')
        super().__init__(**kwargs)


class NCP(Base):
    __filename__ = "Unparished areas names and codes EN as at 12_14.xlsx"
    __tablename__ = "ncp_non_civil_parish_2014"

    ncp_code = Column(String(CODE_LEN), primary_key=True)
    ncp_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'NCP14CD', 'ncp_code')
        rename_kwarg(kwargs, 'NCP14NM', 'ncp_name')
        super().__init__(**kwargs)


class WestminsterConstituency(Base):
    __filename__ = "Westminster Parliamentary Constituency names and codes " \
                   "UK as at 12_14.xlsx"
    __tablename__ = "pcon_westminster_parliamentary_constituency_2014"

    pcon_code = Column(String(CODE_LEN), primary_key=True)
    pcon_name = Column(String(NAME_LEN))

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'PCON14CD', 'pcon_code')
        rename_kwarg(kwargs, 'PCON14NM', 'pcon_name')
        super().__init__(**kwargs)


# =============================================================================
# Models: centroids
# =============================================================================
# http://webarchive.nationalarchives.gov.uk/20160105160709/http://www.ons.gov.uk/ons/guide-method/geography/products/census/spatial/centroids/index.html
#
# Looking at lower_layer_super_output_areas_(e+w)_2011_population_weighted_centroids_v2.zip : # noqa
# - LSOA_2011_EW_PWC.shp -- probably a Shape file;
#   ... yes
#   ... https://en.wikipedia.org/wiki/Shapefile
#   ... ... describes most of the other files
# - LSOA_2011_EW_PWC_COORD_V2.CSV  -- LSOA to centroid coordinates

class PopWeightedCentroidsLsoa2011(Base):
    __filename__ = "LSOA_2011_EW_PWC_COORD_V2.CSV"
    __tablename__ = "pop_weighted_centroids_lsoa_2011"
    # __debug_content__ = True

    lsoa_code = Column(String(CODE_LEN), primary_key=True)
    lsoa_name = Column(String(NAME_LEN))
    bng_north = Column(Integer, doc="British National Grid, North (m)")
    bng_east = Column(Integer, doc="British National Grid, East (m)")
    # https://en.wikipedia.org/wiki/Ordnance_Survey_National_Grid#All-numeric_grid_references  # noqa
    latitude = Column(Numeric(precision=13, scale=10),
                      doc="Latitude (degrees, 10dp)")
    longitude = Column(Numeric(precision=13, scale=10),
                       doc="Longitude (degrees, 10dp)")
    # ... there are some with 10dp, e.g. 0.0000570995
    # ... (precision - scale) = number of digits before '.'
    # ... which can't be more than 3 for any latitude/longitude

    def __init__(self, **kwargs: Any) -> None:
        rename_kwarg(kwargs, 'LSOA11CD', 'lsoa_code')
        rename_kwarg(kwargs, 'LSOA11NM', 'lsoa_name')
        rename_kwarg(kwargs, 'BNGNORTH', 'bng_north')
        rename_kwarg(kwargs, 'BNGEAST', 'bng_east')
        rename_kwarg(kwargs, 'LONGITUDE', 'longitude')
        rename_kwarg(kwargs, 'LATITUDE', 'latitude')
        # MySQL doesn't care if you pass a string to a numeric field, but
        # SQL server does. So:
        convert_int(kwargs, 'bng_north')
        convert_int(kwargs, 'bng_east')
        convert_float(kwargs, 'longitude')
        convert_float(kwargs, 'latitude')
        super().__init__(**kwargs)
        if not self.lsoa_code:
            raise ValueError("Can't have a blank lsoa_code")


# =============================================================================
# Files -> table data
# =============================================================================

def populate_postcode_table(filename: str,
                            session: Session,
                            args: Any,
                            commit: bool = True) -> None:
    tablename = Postcode.__tablename__
    table = Postcode.__table__
    if not args.replace:
        engine = session.bind
        if engine.has_table(tablename):
            log.info("Table {} exists; skipping".format(tablename))
            return
    log.info("Dropping/recreating table: {}".format(tablename))
    table.drop(checkfirst=True)
    table.create(checkfirst=True)
    log.info("Using ONSPD data file: {}".format(filename))
    n = 0
    n_inserted = 0
    extra_fields = []
    db_fields = sorted(k for k in Postcode.__table__.columns.keys()
                       if k != 'pcd_nospace')
    with open(filename) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            n += 1
            if n % args.reportevery == 0:
                log.info("Processing row {}: {} ({} inserted)".format(
                    n, row['pcds'], n_inserted))
                # log.debug(row)
            if n == 1:
                file_fields = sorted(row.keys())
                missing_fields = sorted(set(db_fields) - set(file_fields))
                extra_fields = sorted(set(file_fields) - set(db_fields))
                if missing_fields:
                    log.warning("Fields in database but not file: {}".format
                                (missing_fields))
                if extra_fields:
                    log.warning("Fields in file but not database : {}".format(
                        extra_fields))
            for k in extra_fields:
                del row[k]
            if args.startswith:
                ok = False
                for s in args.startswith:
                    if row['pcd'].startswith(s):
                        ok = True
                        break
                if not ok:
                    continue
            obj = Postcode(**row)
            session.add(obj)
            n_inserted += 1
            if commit and n % args.commitevery == 0:
                commit_and_announce(session)
    if commit:
        commit_and_announce(session)


# BASETYPE = TypeVar('BASETYPE', bound=Base)
# http://mypy.readthedocs.io/en/latest/kinds_of_types.html
# https://docs.python.org/3/library/typing.html


def populate_generic_lookup_table(sa_class: Any,  # Type[Base], Type[BASETYPE] fail on methods  # noqa
                                  datadir: str,
                                  session: Session,
                                  args: Any,
                                  commit: bool = True) -> None:
    """
    The .TXT files look at first glance like tab-separated values files,
    but in some cases have inconsistent numbers of tabs (e.g. "2011 Census
    Output Area Classification Names and Codes UK.txt"). So we'll use the
    .XLSX files.

    If the headings parameter is passed, those headings are used. Otherwise,
    the first row is used for headings.
    """
    tablename = sa_class.__tablename__
    filename = find_first(sa_class.__filename__, datadir)
    headings = getattr(sa_class, '__headings__', [])
    debug = getattr(sa_class, '__debug_content__', False)
    n = 0

    if not args.replace:
        engine = session.bind
        if engine.has_table(tablename):
            log.info("Table {} exists; skipping".format(tablename))
            return

    log.info("Dropping/recreating table: {}".format(tablename))
    sa_class.__table__.drop(checkfirst=True)
    sa_class.__table__.create(checkfirst=True)

    log.info('Processing file "{}" -> table "{}"'.format(filename, tablename))
    ext = os.path.splitext(filename)[1].lower()
    type_xlsx = ext in ['.xlsx']
    type_csv = ext in ['.csv']
    file = None  # type: TextIO

    def dict_from_rows(row_iterator: Iterable[List]) \
            -> Generator[Dict, None, None]:
        local_headings = headings
        first_row = True
        for row in row_iterator:
            values = values_from_row(row)
            if first_row and not local_headings:
                local_headings = values
            else:
                yield dict(zip(local_headings, values))
            first_row = False

    if type_xlsx:
        workbook = openpyxl.load_workbook(filename)  # read_only=True
        # openpyxl BUG: with read_only=True, cells can have None as their value
        # when they're fine if opened in non-read-only mode.
        # May be related to this:
        # https://bitbucket.org/openpyxl/openpyxl/issues/601/read_only-cell-row-column-attributes-are  # noqa
        sheet = workbook.active
        dict_iterator = dict_from_rows(sheet.iter_rows())
    elif type_csv:
        file = open(filename, 'r')
        csv_reader = csv.DictReader(file)
        dict_iterator = csv_reader
    else:
        workbook = xlrd.open_workbook(filename)
        sheet = workbook.sheet_by_index(0)
        dict_iterator = dict_from_rows(sheet.get_rows())
    for datadict in dict_iterator:
        n += 1
        if debug:
            log.critical("{}: {}".format(n, datadict))
        # filter out blanks:
        datadict = {k: v for k, v in datadict.items() if k}
        obj = sa_class(**datadict)
        session.add(obj)
        if commit and n % args.commitevery == 0:
            commit_and_announce(session)
    if commit:
        commit_and_announce(session)
    log.info("... inserted {} rows".format(n))

    if file:
        file.close()


# =============================================================================
# Docs
# =============================================================================

def show_docs() -> None:
    columns = sorted(Postcode.__table__.columns.keys())
    pt = prettytable.PrettyTable(
        ["postcode field", "Description"],
        # header=False,
        border=True,
        hrules=prettytable.ALL,
        vrules=prettytable.NONE,
    )
    pt.align = 'l'
    pt.valign = 't'
    pt.max_width = 80
    for col in columns:
        doc = getattr(Postcode, col).doc
        doc = docx_process_simple_text(doc, width=70)
        ptrow = [col, doc]
        pt.add_row(ptrow)
    print(pt.get_string())


# =============================================================================
# Main
# =============================================================================

def main() -> None:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
        r"""
-   This program reads data from the UK Office of National Statistics Postcode
    Database (ONSPD) and inserts it into a database.

-   You will need to download the ONSPD from
        https://geoportal.statistics.gov.uk/geoportal/catalog/content/filelist.page
    e.g. ONSPD_MAY_2016_csv.zip (79 Mb), and unzip it (>1.4 Gb) to a directory.
    Tell this program which directory you used.

-   Specify your database as an SQLAlchemy connection URL: see
        http://docs.sqlalchemy.org/en/latest/core/engines.html
    The general format is:
        dialect[+driver]://username:password@host[:port]/database[?key=value...]

-   If you get an error like:
        UnicodeEncodeError: 'latin-1' codec can't encode character '\u2019' in
        position 33: ordinal not in range(256)
    then try appending "?charset=utf8" to the connection URL.

-   ONS POSTCODE DATABASE LICENSE.
    Output using this program must add the following attribution statements:

    Contains OS data © Crown copyright and database right [year]
    Contains Royal Mail data © Royal Mail copyright and database right [year]
    Contains National Statistics data © Crown copyright and database right [year]

    See http://www.ons.gov.uk/methodology/geography/licences
    """)  # noqa
    parser.add_argument(
        "--dir", default=DEFAULT_ONSPD_DIR,
        help="Root directory of unzipped ONSPD download (default: {})".format(
            DEFAULT_ONSPD_DIR))
    parser.add_argument(
        "--url", help="SQLAlchemy database URL")
    parser.add_argument(
        "--echo", action="store_true", help="Echo SQL")
    parser.add_argument(
        "--reportevery", type=int, default=DEFAULT_REPORT_EVERY,
        help="Report every n rows (default: {})".format(DEFAULT_REPORT_EVERY))
    parser.add_argument(
        "--commitevery", type=int, default=DEFAULT_COMMIT_EVERY,
        help=(
            "Commit every n rows (default: {}). If you make this too large "
            "(relative e.g. to your MySQL max_allowed_packet setting, you may"
            " get crashes with errors like 'MySQL has gone away'.".format(
                DEFAULT_COMMIT_EVERY)))
    parser.add_argument(
        "--startswith", nargs="+",
        help="Restrict to postcodes that start with one of these strings")
    parser.add_argument(
        "--replace", action="store_true",
        help="Replace tables even if they exist (default: skip existing "
             "tables)")
    parser.add_argument(
        "--skiplookup", action="store_true",
        help="Skip generation of code lookup tables")
    parser.add_argument(
        "--specific_lookup_tables", nargs="*",
        help="Within the lookup tables, process only specific named tables")
    parser.add_argument(
        "--list_lookup_tables", action="store_true",
        help="List all possible lookup tables, then stop")
    parser.add_argument(
        "--skippostcodes", action="store_true",
        help="Skip generation of main (large) postcode table")
    parser.add_argument(
        "--docsonly", action="store_true",
        help="Show help for postcode table then stop")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Verbose")
    args = parser.parse_args()
    rootlogger = logging.getLogger()
    configure_logger_for_colour(
        rootlogger, level=logging.DEBUG if args.verbose else logging.INFO)
    log.debug("args = {}".format(repr(args)))

    if args.docsonly:
        show_docs()
        sys.exit(0)

    classlist = [
        # Core lookup tables:
        # In alphabetical order of filename:
        OAClassification,
        BUA,
        BUASD,
        CASWard,
        CCG,
        CHP,
        Country,
        County2010,
        County1991,
        Datazone,
        DCELLS,
        District,
        EER,
        EnterpriseRegion,
        HealthAuthority,
        HealthBoardNI,
        HealthBoardSC,
        HSCBNI,
        IMDLookupEN,
        IMDLookupNI,
        IMDLookupSC,
        IMDLookupWA,
        IZ2005,
        IZ2011,
        LAU,
        LAD,
        LCG,
        LEA,
        LEP,
        LHB2006,
        LLSC,
        LHB2014,
        LSOAEW2004,
        LSOANI2005,
        LSOAEW2011,
        MSOAEW2004,
        MSOAEW2011,
        NationalPark,
        PanSHA,
        Parish,
        PCT2011,
        PCT2005,
        PFA,
        GOR,
        SHA2004,
        SHA2010,
        SSR,
        Ward2005,
        TTWA,
        NCP,
        Ward2016,
        Ward1998,
        Ward1991,
        WestminsterConstituency,
        # Centroids:
        PopWeightedCentroidsLsoa2011,
    ]

    if args.list_lookup_tables:
        tables_files = []
        for sa_class in classlist:
            tables_files.append((sa_class.__tablename__,
                                 sa_class.__filename__))
        tables_files.sort(key=lambda x: x[0])
        for table, file in tables_files:
            print("Table {} from file {}".format(table, repr(file)))
        return

    if not args.url:
        print("Must specify URL")
        return

    engine = create_engine(args.url, echo=args.echo, encoding=CHARSET)
    metadata.bind = engine
    session = sessionmaker(bind=engine)()

    log.info("Using directory: {}".format(args.dir))
    # lookupdir = os.path.join(args.dir, "Documents")
    lookupdir = args.dir
    # datadir = os.path.join(args.dir, "Data")
    datadir = args.dir

    if not args.skiplookup:
        for sa_class in classlist:
            if (args.specific_lookup_tables and
                    sa_class.__tablename__ not in args.specific_lookup_tables):
                continue
            if (sa_class.__tablename__ ==
                    "ccg_clinical_commissioning_group_england_2015"):
                log.warning("Ignore warning 'Discarded range with reserved "
                            "name' below; it works regardless")
            populate_generic_lookup_table(sa_class, lookupdir, session, args)
    if not args.skippostcodes:
        populate_postcode_table(find_first("ONSPD_*.csv", datadir),
                                session, args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/nlp_manager/regex_test.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import typing.re
from typing import List, Tuple
from crate_anon.nlp_manager.regex_func import compile_regex


def f_score(precision: float, recall: float, beta: float = 1) -> float:
    # https://en.wikipedia.org/wiki/F1_score
    beta_sq = beta ** 2
    return (
        (1 + beta_sq) * precision * recall / ((beta_sq * precision) + recall)
    )


def get_compiled_regex_results(compiled_regex: typing.re.Pattern,
                               text: str) -> List[str]:
    results = []  # type: List[str]
    for m in compiled_regex.finditer(text):
        results.append(m.group(0))
    return results


def print_compiled_regex_results(compiled_regex: typing.re.Pattern, text: str,
                                 prefix_spaces: int = 4) -> None:
    results = get_compiled_regex_results(compiled_regex, text)
    print("{}{} -> {}".format(' ' * prefix_spaces,
                              repr(text), repr(results)))


def test_text_regex(name: str,
                    regex_text: str,
                    test_expected_list: List[Tuple[str, List[str]]],
                    verbose: bool = False) -> None:
    print("Testing regex named {}".format(name))
    compiled_regex = compile_regex(regex_text)
    if verbose:
        print("... regex text:\n{}".format(regex_text))
    for test_string, expected_values in test_expected_list:
        actual_values = get_compiled_regex_results(compiled_regex, test_string)
        assert actual_values == expected_values, (
            "Regex {name}: Expected {expected_values}, got {actual_values}, "
            "when parsing {test_string}. Regex text:\n{regex_text}]".format(
                name=name,
                expected_values=expected_values,
                actual_values=actual_values,
                test_string=repr(test_string),
                regex_text=regex_text,
            )
        )
    print("... OK")
    # print_compiled_regex_results(compiled_regex, text,
    #                              prefix_spaces=prefix_spaces)



#!/usr/bin/env python
# crate_anon/crateweb/research/sql_grammar.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Two approaches:

- sqlparse
  https://sqlparse.readthedocs.io/en/latest/
  ... LESS GOOD.

- pyparsing
  https://pyparsing.wikispaces.com/file/view/select_parser.py
  http://pyparsing.wikispaces.com/file/view/simpleSQL.py
  http://sacharya.com/parsing-sql-with-pyparsing/
  http://bazaar.launchpad.net/~eleybourn/temporal-proxy/devel/view/head:/parser.py  # noqa
  http://pyparsing.wikispaces.com/file/view/select_parser.py/158651233/select_parser.py  # noqa

  - Maximum recursion depth exceeded:
    http://pyparsing.wikispaces.com/share/view/11262229

  - Match-first (|) versus match-longest (^):
    http://pyparsing.wikispaces.com/HowToUsePyparsing

  - How to do arithmetic recursive parsing:
    http://stackoverflow.com/questions/1345039

  - Somebody else's MySQL parser:
    https://github.com/gburns/mysql_pyparsing/blob/master/src/sqlparser.py

  - ... or generic SQL:
    http://pyparsing.wikispaces.com/file/view/select_parser.py/158651233/select_parser.py  # noqa

ANSI SQL syntax:
    http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
    ... particularly the formal specifications in chapter 5 on wards
"""

import logging
import re
from typing import List, Union

from pyparsing import (
    Combine,
    Literal,
    Optional,
    ParseException,
    ParserElement,
    ParseResults,
    QuotedString,
    Regex,
    Suppress,
    Token,
    restOfLine,
    ZeroOrMore,
)
import sqlparse

from crate_anon.common.logsupport import main_only_quicksetup_rootlogger

log = logging.getLogger(__name__)


# =============================================================================
# pyparsing helpers
# =============================================================================

def delim_list(expr_: Token,
               delim: str = ",",
               combine: bool = False,
               combine_adjacent: bool = False,
               suppress_delim: bool = False):
    # A better version of delimitedList
    # BUT SEE ALSO ALTERNATIVE: http://stackoverflow.com/questions/37926516
    name = str(expr_) + " [" + str(delim) + " " + str(expr_) + "]..."
    if combine:
        return Combine(expr_ + ZeroOrMore(delim + expr_),
                       adjacent=combine_adjacent).setName(name)
    else:
        if suppress_delim:
            return (expr_ + ZeroOrMore(Suppress(delim) + expr_)).setName(name)
        else:
            return (expr_ + ZeroOrMore(delim + expr_)).setName(name)


WORD_BOUNDARY = r"\b"
# The meaning of \b:
# http://stackoverflow.com/questions/4213800/is-there-something-like-a-counter-variable-in-regular-expression-replace/4214173#4214173  # noqa


def word_regex_element(word: str) -> str:
    return WORD_BOUNDARY + word + WORD_BOUNDARY


def multiple_words_regex_element(
        words_as_string_or_list: Union[str, List[str]]) -> str:
    if isinstance(words_as_string_or_list, list):
        wordlist = words_as_string_or_list
    elif isinstance(words_as_string_or_list, str):
        wordlist = words_as_string_or_list.split()
    else:
        raise ValueError("Bad wordlist to multiple_words_regex_element")
    return WORD_BOUNDARY + "(" + "|".join(wordlist) + ")" + WORD_BOUNDARY


def make_pyparsing_regex(regex_str: str,
                         caseless: bool = False,
                         name: str = None) -> Regex:
    flags = re.IGNORECASE if caseless else 0
    result = Regex(regex_str, flags=flags)
    if name:
        result.setName(name)
    return result


def make_words_regex(words_as_string_or_list: Union[str, List[str]],
                     caseless: bool = False,
                     name: str = None,
                     debug: bool = False) -> Regex:
    regex_str = multiple_words_regex_element(words_as_string_or_list)
    if debug:
        log.debug(regex_str)
    return make_pyparsing_regex(regex_str, caseless=caseless, name=name)


def make_regex_except_words(word_pattern: str,
                            negative_words_str_or_list: Union[str, List[str]],
                            caseless: bool = False,
                            name: str = None,
                            debug: bool = False) -> Regex:
    # http://regexr.com/
    regex_str = r"(?!{negative}){positive}".format(
        negative=multiple_words_regex_element(negative_words_str_or_list),
        positive=(WORD_BOUNDARY + word_pattern + WORD_BOUNDARY),
    )
    if debug:
        log.debug(regex_str)
    return make_pyparsing_regex(regex_str, caseless=caseless, name=name)


def sql_keyword(word: str) -> Regex:
    regex_str = word_regex_element(word)
    return make_pyparsing_regex(regex_str, caseless=True, name=word)


def bracket(fragment: str) -> str:
    return "(" + fragment + ")"


def single_quote(fragment: str) -> str:
    return "'" + fragment + "'"


# =============================================================================
# Potential keywords
# =============================================================================

# <reserved word>
# http://www.contrib.andrew.cmu.edu/~shadow/sql/sql2bnf.aug92.txt
ANSI92_RESERVED_WORD_LIST = """
    ABSOLUTE ACTION ADD ALL ALLOCATE ALTER AND ANY ARE AS ASC
        ASSERTION AT AUTHORIZATION AVG
    BEGIN BETWEEN BIT BIT_LENGTH BOTH BY
    CASCADE CASCADED CASE CAST CATALOG CHAR CHARACTER CHAR_LENGTH
        CHARACTER_LENGTH CHECK CLOSE COALESCE COLLATE COLLATION
        COLUMN COMMIT CONNECT CONNECTION CONSTRAINT CONSTRAINTS CONTINUE
        CONVERT CORRESPONDING COUNT CREATE CROSS CURRENT
        CURRENT_DATE CURRENT_TIME CURRENT_TIMESTAMP CURRENT_USER CURSOR
    DATE DAY DEALLOCATE DEC DECIMAL DECLARE DEFAULT DEFERRABLE
        DEFERRED DELETE DESC DESCRIBE DESCRIPTOR DIAGNOSTICS
        DISCONNECT DISTINCT DOMAIN DOUBLE DROP
    ELSE END END-EXEC ESCAPE EXCEPT EXCEPTION EXEC EXECUTE EXISTS
        EXTERNAL EXTRACT
    FALSE FETCH FIRST FLOAT FOR FOREIGN FOUND FROM FULL
    GET GLOBAL GO GOTO GRANT GROUP
    HAVING HOUR
    IDENTITY IMMEDIATE IN INDICATOR INITIALLY INNER INPUT INSENSITIVE
        INSERT INT INTEGER INTERSECT INTERVAL INTO IS ISOLATION
    JOIN
    KEY
    LANGUAGE LAST LEADING LEFT LEVEL LIKE LOCAL LOWER
    MATCH MAX MIN MINUTE MODULE MONTH
    NAMES NATIONAL NATURAL NCHAR NEXT NO NOT NULL NULLIF NUMERIC
    OCTET_LENGTH OF ON ONLY OPEN OPTION OR ORDER OUTER OUTPUT OVERLAPS
    PAD PARTIAL POSITION PRECISION PREPARE PRESERVE PRIMARY
    PRIOR PRIVILEGES PROCEDURE PUBLIC
    READ REAL REFERENCES RELATIVE RESTRICT REVOKE RIGHT ROLLBACK ROWS
    SCHEMA SCROLL SECOND SECTION SELECT SESSION SESSION_USER SET
        SIZE SMALLINT SOME SPACE SQL SQLCODE SQLERROR SQLSTATE
        SUBSTRING SUM SYSTEM_USER
    TABLE TEMPORARY THEN TIME TIMESTAMP TIMEZONE_HOUR TIMEZONE_MINUTE
        TO TRAILING TRANSACTION TRANSLATE TRANSLATION TRIM TRUE
    UNION UNIQUE UNKNOWN UPDATE UPPER USAGE USER USING
    VALUE VALUES VARCHAR VARYING VIEW
    WHEN WHENEVER WHERE WITH WORK WRITE
    YEAR
    ZONE
"""

# Keywords, using regexes for speed
ALL = sql_keyword("ALL")
ALTER = sql_keyword("ALTER")
AND = sql_keyword("AND")
ANY = sql_keyword("ANY")
AS = sql_keyword("AS")
ASC = sql_keyword("ASC")
AVG = sql_keyword("AVG")
BETWEEN = sql_keyword("BETWEEN")
BY = sql_keyword("BY")
CASE = sql_keyword("CASE")
CASCADE = sql_keyword("CASCADE")
COLLATE = sql_keyword("COLLATE")
CREATE = sql_keyword("CREATE")
CROSS = sql_keyword("CROSS")
COUNT = sql_keyword("COUNT")
DATE = sql_keyword("DATE")
DATETIME = sql_keyword("DATETIME")
DELETE = sql_keyword("DELETE")
DEFAULT = sql_keyword("DEFAULT")
DESC = sql_keyword("DESC")
DISTINCT = sql_keyword("DISTINCT")
DROP = sql_keyword("DROP")
ELSE = sql_keyword("ELSE")
END = sql_keyword("END")
ESCAPE = sql_keyword("ESCAPE")
EXISTS = sql_keyword("EXISTS")
FALSE = sql_keyword("FALSE")
FIRST = sql_keyword("FIRST")
FOR = sql_keyword("FOR")
FROM = sql_keyword("FROM")
GROUP = sql_keyword("GROUP")
HAVING = sql_keyword("HAVING")
IN = sql_keyword("IN")
INDEX = sql_keyword("INDEX")
INNER = sql_keyword("INNER")
INSERT = sql_keyword("INSERT")
INTERVAL = sql_keyword("INTERVAL")
INTO = sql_keyword("INTO")
IS = sql_keyword("IS")
JOIN = sql_keyword("JOIN")
KEY = sql_keyword("KEY")
LANGUAGE = sql_keyword("LANGUAGE")
LAST = sql_keyword("LAST")
LEFT = sql_keyword("LEFT")
LIKE = sql_keyword("LIKE")
MATCH = sql_keyword("MATCH")
MAX = sql_keyword("MAX")
MIN = sql_keyword("MIN")
NATURAL = sql_keyword("NATURAL")
NOT = sql_keyword("NOT")
NULL = sql_keyword("NULL")
ON = sql_keyword("ON")
ONLY = sql_keyword("ONLY")
OR = sql_keyword("OR")
ORDER = sql_keyword("ORDER")
OUTER = sql_keyword("OUTER")
RESTRICT = sql_keyword("RESTRICT")
RIGHT = sql_keyword("RIGHT")
SELECT = sql_keyword("SELECT")
SET = sql_keyword("SET")
SUM = sql_keyword("SUM")
TABLE = sql_keyword("TABLE")
THEN = sql_keyword("THEN")
TIME = sql_keyword("TIME")
TIMESTAMP = sql_keyword("TIMESTAMP")
TRUE = sql_keyword("TRUE")
UNION = sql_keyword("UNION")
UPDATE = sql_keyword("UPDATE")
USE = sql_keyword("USE")
USING = sql_keyword("USING")
UNKNOWN = sql_keyword("UNKNOWN")
VALUES = sql_keyword("VALUES")
WHEN = sql_keyword("WHEN")
WHERE = sql_keyword("WHERE")
WITH = sql_keyword("WITH")

# =============================================================================
# Punctuation
# =============================================================================

COMMA = Literal(",")
LPAR = Literal("(")
RPAR = Literal(")")

# =============================================================================
# Types of comments
# =============================================================================

ansi_comment = Literal("--") + restOfLine
bash_comment = Literal("#") + restOfLine

# =============================================================================
# Literals
# =============================================================================
# http://dev.mysql.com/doc/refman/5.7/en/literals.html
# http://dev.mysql.com/doc/refman/5.7/en/date-and-time-literals.html

boolean_literal = make_words_regex("TRUE FALSE", name="boolean")
binary_literal = Regex("(b'[01]+')|(0b[01]+)").setName("binary")
hexadecimal_literal = Regex("(X'[0-9a-fA-F]+')|"
                            "(x'[0-9a-fA-F]+')|"
                            "(0x[0-9a-fA-F]+)").setName("hex")
numeric_literal = Regex(r"[+-]?\d+(\.\d*)?([eE][+-]?\d+)?")
integer = Regex(r"[+-]?\d+")  # ... simple integer
sql_keyword_literal = make_words_regex("NULL TRUE FALSE UNKNOWN",
                                       name="boolean")

string_value_singlequote = QuotedString(
    quoteChar="'", escChar="\\", escQuote="''", unquoteResults=False
).setName("single_q_string")
string_value_doublequote = QuotedString(
    quoteChar='"', escChar='\\', escQuote='""', unquoteResults=False
).setName("double_q_string")
string_literal = (
    string_value_singlequote | string_value_doublequote
).setName("string")

# http://dev.mysql.com/doc/refman/5.7/en/date-and-time-literals.html
RE_YEAR = r"([\d]{4}|[\d]{2})"
RE_MONTH = bracket(r"[0][1-9]|[1][0-2]")  # 01 to 12
RE_DAY = bracket(r"[0][0-1]|[1-2][1-9]|[3][0-1]")  # 01 to 31
RE_HOUR = bracket(r"[0-1][0-9]|[2][0-3]")  # 00 to 23
RE_MINUTE = bracket(r"[0-5][0-9]")  # 00 to 59
RE_SECOND = RE_MINUTE
FRACTIONAL_SECONDS = bracket(r"(.[0-9]{1,6})?")  # dot then from 1-6 digits
PUNCTUATION = r"[_,!@#$%&*;\\|:./\^-]"
DATESEP = bracket(r"{}?".format(PUNCTUATION))  # "any punctuation"... or none
DTSEP = bracket(r"[T ]?")  # T or none
TIMESEP = bracket(r"[:]?")  # can also be none

DATE_CORE_REGEX_STR = (
    "({year}{sep}{month}{sep}{day})".format(
        year=RE_YEAR,
        month=RE_MONTH,
        day=RE_DAY,
        sep=DATESEP
    )
)
DATE_LITERAL_REGEX_STR = single_quote(DATE_CORE_REGEX_STR)
TIME_CORE_REGEX_STR = (
    bracket(
        "({hour}{sep}{minute}{sep}{second}{fraction})|"
        "({hour}{sep}{minute})".format(
            hour=RE_HOUR,
            minute=RE_MINUTE,
            second=RE_SECOND,
            fraction=FRACTIONAL_SECONDS,
            sep=TIMESEP
        )
    )
)
TIME_LITERAL_REGEX_STR = single_quote(TIME_CORE_REGEX_STR)
DATETIME_CORE_REGEX_STR = "({date}{sep}{time})".format(
    date=DATE_CORE_REGEX_STR,
    sep=DTSEP,
    time=TIME_CORE_REGEX_STR)
DATETIME_LITERAL_REGEX_STR = single_quote(DATETIME_CORE_REGEX_STR)

date_string = Regex(DATE_LITERAL_REGEX_STR)
time_string = Regex(TIME_LITERAL_REGEX_STR)
datetime_string = Regex(DATETIME_LITERAL_REGEX_STR)
datetime_literal = (
    (Optional(DATE | 'd') + date_string) |
    (Optional(TIME | 't') + time_string) |
    (Optional(TIMESTAMP | 'ts') + datetime_string)
).setName("datetime")

literal_value = (
    numeric_literal |
    boolean_literal |
    binary_literal |
    hexadecimal_literal |
    sql_keyword_literal |
    string_literal |
    datetime_literal
).setName("literal_value")

# https://dev.mysql.com/doc/refman/5.7/en/date-and-time-functions.html#function_date-add  # noqa
time_unit = make_words_regex(
    "MICROSECOND SECOND MINUTE HOUR DAY WEEK MONTH QUARTER YEAR"
    " SECOND_MICROSECOND"
    " MINUTE_MICROSECOND MINUTE_SECOND"
    " HOUR_MICROSECOND HOUR_SECOND HOUR_MINUTE"
    " DAY_MICROSECOND DAY_SECOND DAY_MINUTE DAY_HOUR"
    " YEAR_MONTH",
    caseless=True,
    name="time_unit"
)

# =============================================================================
# sqlparse tests
# =============================================================================

# def parsed_from_text(sql):
#     statements = sqlparse.parse(sql)
#     if not statements:
#         return None
#     first_statement = statements[0]
#     return first_statement


def format_sql(sql: str, reindent: bool = True, indent_width: int = 4) -> str:
    # https://media.readthedocs.org/pdf/sqlparse/latest/sqlparse.pdf
    return sqlparse.format(
        sql,
        reindent=reindent,
        indent_width=indent_width,
        # keyword_case='upper',
    )
    # DON'T use keyword_case='upper'. It converts things like
    # "SELECT ... AS table_name" to "SELECT ... AS TABLE_NAME".


# def formatted_from_parsed(statement):
#     sql = str(statement)
#     return format_sql(sql)


# =============================================================================
# pyparsing test
# =============================================================================

# See also: parser.runTests()

def standardize_for_testing(text: str) -> str:
    text = " ".join(text.split())
    text = text.replace("- ", "-")
    text = text.replace("+ ", "+")
    text = text.replace("~ ", "~")
    text = text.replace("! ", "!")
    text = text.replace(" (", "(")
    text = text.replace("( ", "(")
    text = text.replace(" )", ")")
    return text


def test_succeed(parser: ParserElement,
                 text: str,
                 target: str = None,
                 skip_target: bool = False,
                 show_raw: bool = False,
                 verbose: bool = False) -> None:
    log.critical("Testing to succeed: " + text)
    if target is None:
        target = text
    try:
        p = parser.parseString(text, parseAll=True)
        log.debug("Success: {} -> {}".format(text, text_from_parsed(p)))
        if show_raw:
            log.debug("... raw: {}".format(p))
        if verbose:
            log.debug("... dump:\n{}".format(p.dump()))
    except ParseException as exception:
        log.debug("ParseException on: {}\n... parser: {}".format(text, parser))
        print(statement_and_failure_marker(text, exception))
        raise
    if not skip_target:
        intended = standardize_for_testing(target)
        raw = text_from_parsed(p)
        actual = standardize_for_testing(raw)
        if intended != actual:
            raise ValueError(
                "Failure on: {input}\n"
                "-> Raw output:\n{raw}\n"
                "-> Standardized output:\n{actual}\n"
                "... should have been:\n{intended}\n"
                "... parser: {parser}\n"
                "... as list: {as_list}]".format(
                    input=text,
                    raw=repr(raw),
                    actual=repr(actual),
                    intended=repr(intended),
                    parser=parser,
                    as_list=repr(p.asList())))


def test_fail(parser: ParserElement, text: str, verbose: bool = True) -> None:
    if verbose:
        log.critical("Testing to fail: " + text)
    try:
        p = parser.parseString(text, parseAll=True)
        raise ValueError(
            "Succeeded erroneously (no ParseException): {input}\n"
            "-> structured format: {p}\n"
            "-> Raw text output: {raw}\n"
            "... parser: {parser}]".format(
                input=text,
                raw=text_from_parsed(p),
                p=p,
                parser=parser))
    except ParseException:
        log.debug("Correctly failed: {}".format(text))


def test(parser: ParserElement, text: str) -> None:
    print("STATEMENT:\n{}".format(text))

    # scanned = parser.scanString(text)
    # for tokens, start, end in scanned:
    #     print(start, end, tokens)

    try:
        tokens = parser.parseString(text, parseAll=True)
        print(tokens.asXML())
        # print(tokens.dump())
        # print(text_from_parsed(tokens))
        print("tokens = {}".format(tokens))
        d = tokens.asDict()
        for k, v in d.items():
            print("tokens.{} = {}".format(k, v))
        for c in tokens.columns:
            print("column: {}".format(c))
    except ParseException as err:
        print(statement_and_failure_marker(text, err))
        raise
    print()


def flatten(*args):
    for x in args:
        if isinstance(x, (list, tuple)):
            for y in flatten(*x):
                yield y
        else:
            yield x


def text_from_parsed(parsetree: ParseResults,
                     formatted: bool = True,
                     indent_width: int = 4) -> str:
    nested_list = parsetree.asList()
    flattened_list = flatten(nested_list)
    plain_str = " ".join(flattened_list)
    if not formatted:
        return plain_str

    # Forget my feeble efforts for now (below) and use sqlparse:
    return format_sql(plain_str, reindent=True, indent_width=indent_width)

    # result = ""
    # newline_before = ["SELECT", "FROM", "WHERE", "GROUP"]
    # indent_after = ["SELECT", "FROM", "WHERE", "GROUP"]
    # no_space_before = [",", ")"]
    # no_space_after = ["("]
    # at_line_start = True
    # indent = 0
    #
    # def start_new_line(indent_change=0):
    #     nonlocal at_line_start
    #     nonlocal indent
    #     nonlocal result
    #     if at_line_start:
    #         return
    #     indent += indent_change
    #     result += "\n" + " " * indent * indent_spaces
    #     at_line_start = True
    #
    # previous_item = None
    # for index, item in enumerate(flattened_list):
    #     if item in newline_before:
    #         start_new_line(indent_change=-1)
    #     leading_space = (
    #         not at_line_start and
    #         item not in no_space_before and
    #         previous_item not in no_space_after
    #     )
    #     if leading_space:
    #         result += " "
    #     result += item
    #     at_line_start = False
    #     if item in indent_after:
    #         start_new_line(indent_change=1)
    #     previous_item = item
    # return result


def statement_and_failure_marker(text: str,
                                 parse_exception: ParseException) -> str:
    output_lines = []
    for line_index, line in enumerate(text.split("\n")):
        output_lines.append(line)
        if parse_exception.lineno == line_index + 1:
            output_lines.append("-" * (parse_exception.col - 1) + "^")
    return "\n".join(output_lines)


# =============================================================================
# Base class for SQL grammar
# =============================================================================

class SqlGrammar(object):
    # For type checker:
    bare_identifier_word = None
    column_spec = None
    expr = None
    expr_term = None
    identifier = None
    join_op = None
    join_source = None
    keyword = None
    result_column = None
    select_statement = None
    table_spec = None

    def __init__(self):
        pass

    @classmethod
    def quote_identifier_if_required(cls, identifier: str,
                                     debug_force_quote: bool = False) -> str:
        if debug_force_quote:
            return cls.quote_identifier(identifier)
        if cls.is_quoted(identifier):
            return identifier
        if cls.requires_quoting(identifier):
            return cls.quote_identifier(identifier)
        return identifier

    @classmethod
    def quote_identifier(cls, identifier: str) -> str:
        raise NotImplementedError()

    @classmethod
    def is_quoted(cls, identifier: str) -> bool:
        raise NotImplementedError()

    @classmethod
    def requires_quoting(cls, identifier: str) -> bool:
        raise NotImplementedError()

    @classmethod
    def get_grammar(cls):
        raise NotImplementedError()

    @classmethod
    def get_column_spec(cls):
        raise NotImplementedError()

    @classmethod
    def get_result_column(cls):
        raise NotImplementedError()

    @classmethod
    def get_join_op(cls):
        raise NotImplementedError()

    @classmethod
    def get_table_spec(cls):
        raise NotImplementedError()

    @classmethod
    def get_join_constraint(cls):
        raise NotImplementedError()

    @classmethod
    def get_select_statement(cls):
        raise NotImplementedError()

    @classmethod
    def get_expr(cls):
        raise NotImplementedError()

    @classmethod
    def get_where_clause(cls):
        raise NotImplementedError()

    @classmethod
    def get_where_expr(cls):
        raise NotImplementedError()

    @classmethod
    def test(cls, test_expr: bool = True):
        cls.test_dialect_specific_1()
        cls.test_identifiers()
        if test_expr:
            cls.test_expr()
        cls.test_sql_core()
        cls.test_dialect_specific_2()

    @classmethod
    def test_select(cls, text: str) -> None:
        test_succeed(cls.select_statement, text)

    @classmethod
    def test_select_fail(cls, text: str) -> None:
        test_fail(cls.select_statement, text)

    @classmethod
    def test_identifiers(cls):
        # ---------------------------------------------------------------------
        # Identifiers
        # ---------------------------------------------------------------------
        log.info("Testing keyword")
        # print(cls.keyword.pattern)
        test_succeed(cls.keyword, "TABLE")
        test_fail(cls.keyword, "thingfor")  # shouldn't match FOR
        test_fail(cls.keyword, "forename")  # shouldn't match FOR

        log.info("Testing bare_identifier_word")
        test_succeed(cls.bare_identifier_word, "blah")
        test_fail(cls.bare_identifier_word, "FROM")
        test_succeed(cls.bare_identifier_word, "forename")

        log.info("Testing identifier")
        test_succeed(cls.identifier, "blah")
        test_succeed(cls.identifier, "idx1")
        test_succeed(cls.identifier, "idx2")
        test_succeed(cls.identifier, "a")
        log.info("... done")

        log.info("Testing table_spec")
        test_succeed(cls.table_spec, "mytable")
        test_succeed(cls.table_spec, "mydb.mytable")
        test_fail(cls.table_spec, "mydb . mytable")

        log.info("Testing column_spec")
        test_succeed(cls.column_spec, "mycol")
        test_succeed(cls.column_spec, "forename")
        test_succeed(cls.column_spec, "mytable.mycol")
        test_succeed(cls.column_spec, "t1.a")
        test_succeed(cls.column_spec, "mydb.mytable.mycol")
        test_fail(cls.column_spec, "mydb . mytable . mycol")

    @classmethod
    def test_expr(cls):
        # ---------------------------------------------------------------------
        # Expressions
        # ---------------------------------------------------------------------

        log.info("Testing expr_term")
        test_succeed(cls.expr_term, "5")
        test_succeed(cls.expr_term, "-5")
        test_succeed(cls.expr_term, "5.12")
        test_succeed(cls.expr_term, "'string'")
        test_succeed(cls.expr_term, "mycol")
        test_succeed(cls.expr_term, "myfunc(myvar, 8)")
        test_succeed(cls.expr_term, "INTERVAL 5 MICROSECOND")
        test_succeed(cls.expr_term, "(SELECT 1)")
        test_succeed(cls.expr_term, "(1, 2, 3)")
        # Concatenated rubbish shouldn't pass
        # test_fail(cls.expr_term, "one two three four")  # matches "one"

        log.info("Testing expr")
        test_succeed(cls.expr, "5")
        test_succeed(cls.expr, "-5")
        test_succeed(cls.expr, "a")
        test_succeed(cls.expr, "mycol1 || mycol2")
        test_succeed(cls.expr, "+mycol")
        test_succeed(cls.expr, "-mycol")
        test_succeed(cls.expr, "~mycol")
        test_succeed(cls.expr, "!mycol")
        test_succeed(cls.expr, "a | b")
        test_succeed(cls.expr, "a & b")
        test_succeed(cls.expr, "a << b")
        test_succeed(cls.expr, "a >> b")
        test_succeed(cls.expr, "a + b")
        test_succeed(cls.expr, "a - b")
        test_succeed(cls.expr, "a * b")
        test_succeed(cls.expr, "a / b")
        test_succeed(cls.expr, "a % b")
        test_succeed(cls.expr, "a ^ b")
        test_succeed(cls.expr, "a ^ (b + (c - d) / e)")
        test_succeed(cls.expr, "a NOT IN (SELECT 1)")
        test_succeed(cls.expr, "a IN (1, 2, 3)")
        test_succeed(cls.expr, "a IS NULL")
        test_succeed(cls.expr, "a IS NOT NULL")
        test_fail(cls.expr, "IS NULL")
        test_fail(cls.expr, "IS NOT NULL")
        test_succeed(cls.expr, "(a * (b - 3)) > (d - 2)")

        # Bare keywords shouldn't pass:
        test_fail(cls.expr, "COUNT")

        # Concatenated rubbish shouldn't pass
        test_fail(cls.expr, "one two three four")

    @classmethod
    def test_sql_core(cls):
        log.info("Testing join_op")
        test_succeed(cls.join_op, ",")
        test_succeed(cls.join_op, "INNER JOIN")

        log.info("Testing join_source")
        test_succeed(cls.join_source, "a")
        test_succeed(cls.join_source, "a INNER JOIN b")
        test_succeed(cls.join_source, "a, b")

        log.info("Testing result_column")
        test_succeed(cls.result_column, "t1.a")
        test_succeed(cls.result_column, "col1")
        test_succeed(cls.result_column, "t1.col1")
        test_succeed(cls.result_column, "col1 AS alias")
        test_succeed(cls.result_column, "t1.col1 AS alias")
        # test_succeed(cls.result_column, "*")
        test_succeed(cls.result_column, "COUNT(*)")
        test_succeed(cls.result_column, "COUNT(*) AS alias")
        test_fail(cls.result_column, "COUNT")

        test_succeed(Combine(COUNT + LPAR) + '*', "COUNT(*")
        test_succeed(Combine(COUNT + LPAR) + '*', "COUNT( *")
        test_fail(Combine(COUNT + LPAR) + '*', "COUNT (*")

        test_fail(cls.result_column, "COUNT (*)")  # space disallowed
        test_succeed(cls.result_column, "COUNT(DISTINCT col)")
        test_succeed(cls.result_column, "MAX(col)")
        test_fail(cls.result_column, "MAX (col)")

        log.info("Testing select_statement")
        cls.test_select("SELECT t1.a, t1.b FROM t1 WHERE t1.col1 IN (1, 2, 3)")
        cls.test_select("SELECT a, b FROM c")  # no WHERE
        cls.test_select("SELECT a, b FROM c WHERE d > 5 AND e = 4")
        cls.test_select("SELECT a, b FROM c INNER JOIN f WHERE d > 5 AND e = 4")
        cls.test_select("SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5")

        log.info("Testing SELECT something AS alias")
        cls.test_select("SELECT col1 AS alias FROM t1")
        cls.test_select("SELECT t1.col1 AS alias FROM t1")

        log.info("Testing nested query: IN")
        cls.test_select("SELECT col1 FROM table1 WHERE col2 IN (SELECT col3 FROM table2)")  # noqa

        log.info("Testing COUNT(*)")
        cls.test_select("SELECT col1, COUNT(*) FROM table1 GROUP BY col1")

        log.info("MySQL rejects COUNT (*) and other aggregate functions with "
                 "a space between the function and the opening bracket")
        cls.test_select_fail(
            "SELECT col1, COUNT (*) FROM table1 GROUP BY col1")

        log.info("Testing SELECT with literal")
        cls.test_select("SELECT 'blah' AS some_literal")

        log.info("Testing SELECT result columns")
        test_succeed(cls.get_result_column(), "somecol")
        test_succeed(cls.get_result_column(), "somecol AS alias")
        test_succeed(cls.get_result_column(), "COUNT(*)")
        test_succeed(cls.get_result_column(), "COUNT(*) AS alias")
        test_succeed(cls.get_result_column(), "'blah'")
        test_succeed(cls.get_result_column(), "'blah' AS alias")
        test_succeed(cls.get_result_column(), "NULL")
        test_succeed(cls.get_result_column(), "NULL AS alias")

    @classmethod
    def test_dialect_specific_1(cls):
        pass

    @classmethod
    def test_dialect_specific_2(cls):
        pass


def test_base_elements() -> None:
    # -------------------------------------------------------------------------
    # pyparsing tests
    # -------------------------------------------------------------------------
    log.info("Testing pyparsing elements")
    regexp_for = Regex(r"\bfor\b", flags=re.IGNORECASE)
    test_succeed(regexp_for, "for")
    test_fail(regexp_for, "blandford")
    test_fail(regexp_for, "upfor")
    test_fail(regexp_for, "forename")

    # -------------------------------------------------------------------------
    # Literals
    # -------------------------------------------------------------------------
    log.info("Testing boolean_literal")
    test_succeed(boolean_literal, "TRUE")
    test_succeed(boolean_literal, "FALSE")
    test_fail(boolean_literal, "blah")

    log.info("Testing binary_literal")
    test_succeed(binary_literal, "b'010101'")
    test_succeed(binary_literal, "0b010101")

    log.info("Testing hexadecimal_literal")
    test_succeed(hexadecimal_literal, "X'12fac'")
    test_succeed(hexadecimal_literal, "x'12fac'")
    test_succeed(hexadecimal_literal, "0x12fac")

    log.info("Testing integer")
    test_succeed(integer, "99")
    test_succeed(integer, "-99")

    log.info("Testing numeric_literal")
    test_succeed(numeric_literal, "45")
    test_succeed(numeric_literal, "+45")
    test_succeed(numeric_literal, "-45")
    test_succeed(numeric_literal, "-45E-3")
    test_succeed(numeric_literal, "-45E3")
    test_succeed(numeric_literal, "-45.32")
    test_succeed(numeric_literal, "-45.32E6")

    log.info("Testing string_value_singlequote")
    test_succeed(string_value_singlequote, "'single-quoted string'")
    log.info("Testing string_value_doublequote")
    test_succeed(string_value_doublequote, '"double-quoted string"')
    log.info("Testing string_literal")
    test_succeed(string_literal, "'single-quoted string'")
    test_succeed(string_literal, '"double-quoted string"')

    log.info("Testing date_string")
    test_succeed(date_string, single_quote("2015-04-14"))
    test_succeed(date_string, single_quote("20150414"))
    log.info("Testing time_string")
    test_succeed(time_string, single_quote("15:23"))
    test_succeed(time_string, single_quote("1523"))
    test_succeed(time_string, single_quote("15:23:00"))
    test_succeed(time_string, single_quote("15:23:00.1"))
    test_succeed(time_string, single_quote("15:23:00.123456"))
    log.info("Testing datetime_string")
    test_succeed(datetime_string, single_quote("2015-04-14 15:23:00.123456"))
    test_succeed(datetime_string, single_quote("2015-04-14T15:23:00.123456"))

    log.info("Testing literal")
    test_succeed(literal_value, "NULL")
    test_succeed(literal_value, "99")
    test_succeed(literal_value, "-99")

    log.info("Testing time_unit")
    test_succeed(time_unit, "MICROSECOND")
    test_succeed(time_unit, "year_month")

    # -------------------------------------------------------------------------
    # Identifiers
    # -------------------------------------------------------------------------

    log.info("Testing FOR")
    # print(FOR.pattern)
    test_succeed(FOR, "for")
    test_fail(FOR, "thingfor")  # shouldn't match FOR
    test_fail(FOR, "forename")  # shouldn't match FOR


# =============================================================================
# main
# =============================================================================

def main() -> None:
    # blank = ''
    # p_blank = parsed_from_text(blank)
    # p_sql1 = parsed_from_text(sql1)
    # print(formatted_from_parsed(p_blank))
    # print(formatted_from_parsed(p_sql1))

    # pyparsing_bugtest_delimited_list_combine()

    log.info("TESTING BASE ELEMENTS")
    test_base_elements()
    log.info("ALL TESTS SUCCESSFUL")


if __name__ == '__main__':
    main_only_quicksetup_rootlogger()
    main()



#!/usr/bin/env python
# crate_anon/nlp_manager/parse_gate.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

The pipe encoding (Python -> Java stdin, Java stdout -> Python) is fixed at
UTF-8 here and in the Java code.

"""

import logging
import os
import shlex
import subprocess
import sys
from typing import Any, Dict, Generator, List, Tuple

from cardinal_pythonlib.rnc_lang import chunks
from sqlalchemy import Column, Index, Integer, Text

from crate_anon.nlp_manager.base_nlp_parser import BaseNlpParser
from crate_anon.nlp_manager.constants import (
    MAX_SQL_FIELD_LEN,
    SqlTypeDbIdentifier,
)
from crate_anon.nlp_manager.nlp_definition import NlpDefinition
from crate_anon.nlp_manager.output_user_config import OutputUserConfig
from crate_anon.nlp_manager.text_handling import tsv_pairs_to_dict

log = logging.getLogger(__name__)


# =============================================================================
# Process handling
# =============================================================================
# Have Python host the client process, communicating with stdin/stdout?
#   http://eyalarubas.com/python-subproc-nonblock.html
#   http://stackoverflow.com/questions/2715847/python-read-streaming-input-from-subprocess-communicate  # noqa
# Java process could be a network server.
#   http://docs.oracle.com/javase/tutorial/networking/sockets/clientServer.html
#   http://www.tutorialspoint.com/java/java_networking.htm
# OK, first one works; that's easier.

class Gate(BaseNlpParser):
    """Class controlling an external process, typically our Java interface to
    GATE programs, CrateGatePipeline.java (but it could be any external
    program).

    We send text to it, it parses the text, and it sends us back results, which
    we return as dictionaries. The specific text sought depends on the
    configuration file and the specific GATE program used.

    PROBLEM when attempting to use KConnect (Bio-YODIE): its source code is
    riddled with direct calls to System.out.println().

    POTENTIAL SOLUTIONS
    - named pipes:
        os.mkfifo() - Unix only.
        win32pipe - http://stackoverflow.com/questions/286614
    - ZeroMQ with some sort of security
        - pip install zmq
        - some sort of Java binding (jzmq, jeromq...)
    - redirect stdout in our Java handler
        System.setOut()
        ... yes, that works.
    """
    NAME = "GATE"

    def __init__(self,
                 nlpdef: NlpDefinition,
                 cfgsection: str,
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef, cfgsection=cfgsection, commit=commit)

        if not nlpdef and not cfgsection:
            # Debugging only
            self._max_external_prog_uses = 0
            self._input_terminator = 'input_terminator'
            self._output_terminator = 'output_terminator'
            typepairs = []  # type: List[str]
            self._progenvsection = ''
            progargs = ''
            logtag = ''
        else:
            self._max_external_prog_uses = nlpdef.opt_int(
                cfgsection, 'max_external_prog_uses', default=0)
            self._input_terminator = nlpdef.opt_str(
                cfgsection, 'input_terminator', required=True)
            self._output_terminator = nlpdef.opt_str(
                cfgsection, 'output_terminator', required=True)
            typepairs = nlpdef.opt_strlist(cfgsection, 'outputtypemap',
                                           required=True, lower=False)
            self._progenvsection = nlpdef.opt_str(cfgsection, 'progenvsection')
            progargs = nlpdef.opt_str(cfgsection, 'progargs', required=True)
            logtag = nlpdef.get_logtag() or '.'

        self._outputtypemap = {}  # type: Dict[str, OutputUserConfig]
        self._type_to_tablename = {}  # type: Dict[str, str]
        for c in chunks(typepairs, 2):
            annottype = c[0]
            outputsection = c[1]
            if annottype != annottype.lower():
                raise Exception(
                    "Section {}: annotation types in outputtypemap must be in "
                    "lower case: change {}".format(cfgsection, annottype))
            # log.critical(outputsection)
            c = OutputUserConfig(nlpdef.get_parser(), outputsection)
            self._outputtypemap[annottype] = c
            self._type_to_tablename[annottype] = c.get_tablename()

        if self._progenvsection:
            self._env = nlpdef.get_env_dict(self._progenvsection, os.environ)
        else:
            self._env = os.environ.copy()
        self._env["NLPLOGTAG"] = logtag
        # ... because passing a "-lt" switch with no parameter will make
        # CrateGatePipeline.java complain and stop

        formatted_progargs = progargs.format(**self._env)
        self._progargs = shlex.split(formatted_progargs)

        self._n_uses = 0
        self._pipe_encoding = 'utf8'
        self._p = None  # the subprocess
        self._started = False

        # Sanity checks
        for ty, tn in self._type_to_tablename.items():
            assert len(tn) <= MAX_SQL_FIELD_LEN, (
                "Table name too long (max {} characters)".format(
                    MAX_SQL_FIELD_LEN))

    @classmethod
    def print_info(cls, file=sys.stdout):
        print("NLP class to talk to GATE apps (https://www.gate.ac.uk/).",
              file=file)

    # -------------------------------------------------------------------------
    # External process control
    # -------------------------------------------------------------------------

    def _start(self) -> None:
        """
        Launch the external process.
        """
        if self._started:
            return
        args = self._progargs
        log.info("launching command: {}".format(args))
        self._p = subprocess.Popen(args,
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   # stderr=subprocess.PIPE,
                                   shell=False,
                                   bufsize=1)
        # ... don't ask for stderr to be piped if you don't want it; firstly,
        # there's a risk that if you don't consume it, something hangs, and
        # secondly if you don't consume it, you see it on the console, which is
        # helpful.
        self._started = True

    def _encode_to_subproc_stdin(self, text: str) -> None:
        """Send text to the external program (via its stdin), encoding it in
        the process (typically to UTF-8)."""
        log.debug("SENDING: " + text)
        bytes_ = text.encode(self._pipe_encoding)
        self._p.stdin.write(bytes_)

    def _flush_subproc_stdin(self) -> None:
        """Flushes what we're sending to the external program via its stdin."""
        self._p.stdin.flush()

    def _decode_from_subproc_stdout(self) -> str:
        """Translate what we've received from the external program's stdout,
        from its specific encoding (usually UTF-8) to a Python string."""
        bytes_ = self._p.stdout.readline()
        text = bytes_.decode(self._pipe_encoding)
        log.debug("RECEIVING: " + repr(text))
        return text

    def _finish(self) -> None:
        """
        Close down the external process.
        """
        if not self._started:
            return
        self._p.communicate()  # close p.stdout, wait for the subprocess to exit
        self._started = False

    # -------------------------------------------------------------------------
    # Input processing
    # -------------------------------------------------------------------------

    def parse(self, text: str) -> Generator[Tuple[str, Dict[str, Any]],
                                            None, None]:
        """
        - Send text to the external process, and receive the result.
        - Note that associated data is not passed into this function, and is
          kept in the Python environment, so we can't run into any problems
          with the transfer to/from the Java program garbling important data.
          All we send to the subprocess is the text (and an input_terminator).
          Then, we may receive MULTIPLE sets of data back ("your text contains
          the following 7 people/drug references/whatever"), followed
          eventually by the output_terminator, at which point this set is
          complete.
        """
        self._start()  # ensure started
        # Send
        log.debug("writing: " + text)
        self._encode_to_subproc_stdin(text)
        self._encode_to_subproc_stdin(os.linesep)
        self._encode_to_subproc_stdin(self._input_terminator + os.linesep)
        self._flush_subproc_stdin()  # required in the Python 3 system

        # Receive
        for line in iter(self._decode_from_subproc_stdout,
                         self._output_terminator + os.linesep):
            # ... iterate until the sentinel output_terminator is received
            line = line.rstrip()  # remove trailing newline
            log.debug("stdout received: " + line)
            d = tsv_pairs_to_dict(line)
            log.debug("dictionary received: {}".format(d))
            try:
                annottype = d['_type'].lower()
            except KeyError:
                raise ValueError("_type information not in data received")
            if annottype not in self._type_to_tablename:
                log.warning(
                    "Unknown annotation type, skipping: {}".format(annottype))
                continue
            yield self._type_to_tablename[annottype], d

        self._n_uses += 1
        # Restart subprocess?
        if 0 < self._max_external_prog_uses <= self._n_uses:
            log.info("relaunching app after {} uses".format(self._n_uses))
            self._finish()
            self._start()
            self._n_uses = 0

    # -------------------------------------------------------------------------
    # Test
    # -------------------------------------------------------------------------

    def test(self, verbose: bool=False) -> None:
        """
        Test the send function.
        """
        self.test_parser([
            "Bob Hope visited Seattle.",
            "James Joyce wrote Ulysses."
        ])

    # -------------------------------------------------------------------------
    # Database structure
    # -------------------------------------------------------------------------

    @staticmethod
    def _standard_columns() -> List[Column]:
        return [
            Column('_type', SqlTypeDbIdentifier,
                   doc="GATE annotation type name"),
            Column('_id', Integer,
                   doc="GATE annotation ID (not clear this is very useful)"),
            Column('_start', Integer,
                   doc="Start position in the content"),
            Column('_end', Integer,
                   doc="End position in the content"),
            Column('_content', Text,
                   doc="Full content marked as relevant."),
        ]

    def dest_tables_columns(self) -> Dict[str, List[Column]]:
        tables = {}  # type: Dict[str, List[Column]]
        for anottype, otconfig in self._outputtypemap.items():
            tables[otconfig.get_tablename()] = (
                self._standard_columns() +
                otconfig.get_columns(self.get_engine())
            )
        return tables

    def dest_tables_indexes(self) -> Dict[str, List[Index]]:
        tables = {}  # type: Dict[str, List[Index]]
        for anottype, otconfig in self._outputtypemap.items():
            tables[otconfig.get_tablename()] = otconfig.get_indexes()
        return tables



#!/usr/bin/env python
# crate_anon/crateweb/research/context_processors.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from typing import Any, Dict

from django.conf import settings
from django.http.request import HttpRequest
# from crate_anon.crateweb.research.models import Query
# from crate_anon.crateweb.extra.request_cache import get_request_cache


# noinspection PyUnusedLocal
def common_context(request: HttpRequest) -> Dict[str, Any]:
    return {
        'RESEARCH_DB_TITLE': settings.RESEARCH_DB_TITLE,
        'nav_on_main_menu': False,
    }
    # Try to minimize SQL here, as these calls will be used for EVERY
    # request.
    # This problem can partially be circumvented with a per-request cache; see
    # http://stackoverflow.com/questions/3151469/per-request-cache-in-django
    # But good practice is: keep queries to a minimum.
